"""
Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE

Contains large portions of
https://github.com/Rapptz/discord.py/blob/master/discord/ext/commands/core.py
The appropriate license is in LICENSE
"""

from __future__ import annotations

import asyncio
import functools
import inspect
import itertools
from collections.abc import Callable, Coroutine, Iterable, Sequence
from time import time
from types import UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Concatenate,
    ForwardRef,
    Generic,
    Literal,
    Protocol,
    TypeAlias,
    Union,
    cast,
    get_args,
    get_origin,
    overload,
)

import typing_extensions
from typing_extensions import ParamSpec, Self, TypedDict, TypeVar, Unpack

from ...channel import UserChannel
from ...chat import Chat
from ...utils import cached_property, maybe_coroutine
from . import converters
from .context import Context
from .cooldown import BucketType, Cooldown
from .errors import *
from .utils import CaseInsensitiveDict, Coro

if TYPE_CHECKING:
    from .bot import Bot
    from .cog import Cog

__all__ = (
    "Command",
    "command",
    "Group",
    "group",
    "check",
    "is_owner",
    "cooldown",
)

P = ParamSpec("P", default=...)
R = TypeVar("R", default=Any, covariant=True)


MaybeCoroFunc: TypeAlias = Callable[P, R | Coro[R]]
CoroFunc: TypeAlias = Callable[P, Coro[R]]
CheckType: TypeAlias = MaybeCoroFunc[["Context"], bool]
ErrT = TypeVar("ErrT", bound=CoroFunc[["Context", Exception], None] | CoroFunc[["Cog", "Context", Exception], None])
InvokeT = TypeVar("InvokeT", bound=CoroFunc[["Context"], None] | CoroFunc[["Cog", "Context"], None])
MaybeCommandT = TypeVar("MaybeCommandT", bound="Callable[..., Command] | Command")
CoroFuncT = TypeVar("CoroFuncT", bound=CoroFunc)


class CommandDeco(Protocol):
    def __call__(self, command: MaybeCommandT, /) -> MaybeCommandT: ...


MaybeBool = TypeVar("MaybeBool", bound=bool | Coro[bool], default=bool | Coro[bool], covariant=True)


class Check(Protocol[MaybeBool]):
    predicate: Callable[[Context[Any]], Coro[bool]]

    @overload
    def __call__(self, context: Context[Any], /) -> MaybeBool: ...

    @overload
    def __call__(self, command: MaybeCommandT, /) -> MaybeCommandT: ...


@converters.converter
def to_bool(argument: str) -> bool:
    lowered = argument.lower()
    if lowered in {"yes", "y", "true", "t", "1", "enable", "on"}:
        return True
    elif lowered in {"no", "n", "false", "f", "0", "disable", "off"}:
        return False
    raise BadArgument(f"{argument!r} is not a recognised boolean option")


class CommandKwargsNoCls(TypedDict, total=False):
    name: str | None
    help: str
    brief: str
    usage: str
    description: str
    aliases: Iterable[str]
    checks: list[Check]
    cooldown: list[Cooldown]
    special_converters: list[converters.Converters]
    cog: Cog
    parent: GroupMixin | Group[Any]
    enabled: bool
    hidden: bool
    case_insensitive: bool


C = TypeVar("C", bound="Command", default="Command")


class CommandKwargs(CommandKwargsNoCls, Generic[C], total=False):
    cls: type[C] | None


CogT = TypeVar("CogT", bound="Cog | Bot | None", default="Cog | Bot | None", covariant=True)


class Command(Generic[CogT, P, R]):
    """A class to represent a command.

    Attributes
    ----------
    name
        The command's name.
    help
        The command's help docstring.
    checks
        A list of the command's checks.
    cooldown
        The command's cooldowns.
    special_converters
        A list of the command's special converters as registered by :meth:`steam.ext.commands.Converter.register`.
    enabled
        Whether the command is enabled.
    cog
        The command's cog.
    parent
        The command's parent.
    description
        The command's description.
    hidden
        Whether the command is hidden.
    aliases
        The command's aliases.
    params
        The command's parameters.
    """

    __original_kwargs__: dict[str, Any]

    def __new__(cls, *args: Any, **kwargs: Any) -> Self:
        self = super().__new__(cls)
        self.__original_kwargs__ = kwargs.copy()
        return self

    def __init__(
        self,
        func: CoroFunc[Concatenate[CogT, Context[Any], P], R] | CoroFunc[Concatenate[Context[Any], P], R],
        /,
        **kwargs: Unpack[CommandKwargs],
    ):
        self.name = kwargs.get("name") or func.__name__
        self.callback = func

        help_doc = kwargs.get("help")
        self.help: str | None = inspect.cleandoc(help_doc) if help_doc is not None else inspect.getdoc(func)

        try:
            checks = func.__commands_checks__  # type: ignore
            checks.reverse()
        except AttributeError:
            checks = kwargs.get("checks", [])
        self.checks: list[Check] = checks

        try:
            cooldown = func.__commands_cooldown__  # type: ignore
        except AttributeError:
            cooldown = kwargs.get("cooldown", [])
        self.cooldown: list[Cooldown] = cooldown

        try:
            special_converters: list[converters.Converters] = func.__special_converters__  # type: ignore
        except AttributeError:
            special_converters = kwargs.get("special_converters", [])
        self.special_converters = special_converters

        self.enabled = kwargs.get("enabled", True)
        self.brief = kwargs.get("brief")
        self.usage = kwargs.get("usage")
        self.cog: Cog | Bot | None = None
        self.parent: GroupMixin | Group[CogT] | None = kwargs.get("parent")
        self.description: str = inspect.cleandoc(kwargs.get("description", ""))
        self.hidden: bool = kwargs.get("hidden", False)
        self.aliases: Sequence[str] = list(kwargs.get("aliases", ()))

        self._before_hook = None
        self._after_hook = None

    def __str__(self) -> str:
        return self.qualified_name

    # @property
    # @overload
    # def callback(
    #     self: Command[None, Concatenate[Context[Any], P], R]
    # ) -> Callable[Concatenate[Context[Any], P], Coro[R]]:
    #     ...

    # @property
    # @overload
    # def callback(
    #     self: Command[CogT, Concatenate[CogT, Context[Any], P], R]
    # ) -> Callable[Concatenate[CogT, Context[Any], P], Coro[R]]:
    #     ...

    @property
    def callback(
        self,
    ) -> Callable[Concatenate[CogT, Context[Any], P], Coro[R]] | Callable[Concatenate[Context[Any], P], Coro[R]]:
        """The internal callback the command holds."""
        return self._callback

    @callback.setter
    def callback(
        self,
        function: (
            Callable[Concatenate[CogT, Context[Any], P], Coro[R]] | Callable[Concatenate[Context[Any], P], Coro[R]]
        ),
    ) -> None:
        if not inspect.iscoroutinefunction(function):
            raise TypeError(f"The callback for the command {function.__name__!r} must be a coroutine function.")

        annotations = typing_extensions.get_type_hints(function)
        for name, annotation in annotations.items():
            if get_origin(annotation) is converters.Greedy and isinstance(annotation.converter, ForwardRef):
                annotations[name] = converters.Greedy[eval(annotation.converter.__forward_code__, function.__globals__)]

        if inspect.ismethod(function):
            function.__func__.__annotations__ = annotations
        else:
            function.__annotations__ = annotations
        self.params = dict(inspect.signature(function, eval_str=True).parameters)
        if not self.params:
            raise TypeError(f'Callback for {self.name} command is missing a "ctx" parameter.') from None

        self.module = function.__module__
        self._callback = function

    @cached_property
    def clean_params(self) -> dict[str, inspect.Parameter]:
        """The command's parameters without ``"self"`` and ``"ctx"``."""
        params = iter(self.params.items())
        if self.cog is not None:
            try:
                next(params)  # cog's "self" param
            except StopIteration:
                raise TypeError(f'Callback for {self.name} command is missing a "self" parameter.') from None
        try:
            next(params)  # context param
        except StopIteration:
            raise TypeError(f'Callback for {self.name} command is missing a "ctx" parameter.') from None
        return dict(params)

    @property
    def qualified_name(self) -> str:
        """The full name of the command, this takes into account subcommands."""
        return " ".join(c.name for c in reversed(self.parents))

    @property
    def parents(self) -> list[Self]:
        """The command's parents."""
        commands: list[Self] = []
        command = self
        while isinstance(command, Command):
            commands.append(command)
            command = command.parent

        return commands

    async def __call__(self, ctx: Context, *args: P.args, **kwargs: P.kwargs) -> R:
        """Calls the internal callback that the command holds.

        Note
        ----
        This bypasses all mechanisms -- including checks, converters, invoke hooks, cooldowns, etc. You must take
        care to pass the proper arguments (excluding the parameter "self" in a cog context) and types to this
        function.
        """
        if self.cog is not None:
            return await self.callback(self.cog, ctx, *args, **kwargs)  # type: ignore
        else:
            return await self.callback(ctx, *args, **kwargs)  # type: ignore

    def error(self, coro: ErrT) -> ErrT:
        """A decorator to register a :term:`coroutine function` to handle a commands ``on_error`` functionality
        similarly to :meth:`steam.ext.commands.Bot.on_command_error`.

        Example:

        .. code:: python

            @bot.command
            async def raise_an_error(ctx: commands.Context) -> None:
                raise Exception("oh no an error")


            @raise_an_error.error
            async def on_error(ctx: commands.Context, error: Exception) -> None:
                await ctx.send(f"{ctx.command.name} raised an exception {error!r}")
        """

        if not inspect.iscoroutinefunction(coro):
            raise TypeError(f"Error handler for {self.name} must be a coroutine function")
        self.on_error = coro
        return coro

    def before_invoke(self, coro: InvokeT) -> InvokeT:
        """Register a :term:`coroutine function` to be run before any arguments are parsed."""

        if not inspect.iscoroutinefunction(coro):
            raise TypeError(f"Hook for {self.name} must be a coroutine function")
        self._before_hook = coro
        return coro

    def after_invoke(self, coro: InvokeT) -> InvokeT:
        """Register a :term:`coroutine function` to be run after the command has been invoked."""

        if not inspect.iscoroutinefunction(coro):
            raise TypeError(f"Hook for {self.name} must be a coroutine function")
        self._after_hook = coro
        return coro

    async def invoke(self, ctx: Context) -> None:
        """Invoke the callback the command holds.

        Parameters
        ----------
        ctx
            The invocation context.
        """
        try:
            if not self.enabled:
                raise CommandDisabled(self)
            for check in itertools.chain(ctx.bot.checks, self.checks):
                if not await check.predicate(ctx):
                    raise CheckFailure("You failed to pass one of the checks for this command")
            for cooldown in self.cooldown:
                cooldown(ctx)
            await self._parse_arguments(ctx)
            await self._call_before_invoke(ctx)
            assert ctx.args is not None
            assert ctx.kwargs is not None
            await self(ctx, *ctx.args, **ctx.kwargs)
        except asyncio.CancelledError:
            return
        except Exception:
            ctx.command_failed = True
            raise
        finally:
            await self._call_after_invoke(ctx)

    async def can_run(self, ctx: Context) -> bool:
        """Whether the command can be run.

        Parameters
        ----------
        ctx
            The invocation context.
        """
        if not self.enabled:
            return False
        for check in self.checks:
            if not await maybe_coroutine(check, ctx):
                return False
        for cooldown in self.cooldown:
            bucket = cooldown.bucket.get_bucket(ctx)
            if cooldown.get_retry_after(bucket, time()):
                return False

        return True

    async def _call_before_invoke(self, ctx: Context) -> None:
        if self._before_hook is not None:
            if self.cog is not None:
                await self._before_hook(self.cog, ctx)  # type: ignore
            else:
                await self._before_hook(ctx)  # type: ignore
        if ctx.bot._before_hook is not None:
            await ctx.bot._before_hook(ctx)

    async def _call_after_invoke(self, ctx: Context) -> None:
        if self._after_hook is not None:
            if self.cog is not None:
                await self._after_hook(self.cog, ctx)  # type: ignore
            else:
                await self._after_hook(ctx)  # type: ignore
        if ctx.bot._after_hook is not None:
            await ctx.bot._after_hook(ctx)

    async def _parse_positional_or_keyword_argument(
        self, ctx: Context, param: inspect.Parameter, args: list[Any]
    ) -> None:
        is_greedy = get_origin(param.annotation) is converters.Greedy
        greedy_args: list[Any] = []
        if ctx.lex.position == ctx.lex.end:
            args.append(await self._get_default(ctx, param))
        for argument in ctx.lex:
            try:
                transformed = await self._transform(ctx, param, argument)
            except BadArgument:
                if not is_greedy:
                    raise
                ctx.lex.undo()  # undo last read string for the next argument
                args.append(tuple(greedy_args))
                break
            if not is_greedy:
                args.append(transformed)
                break
            greedy_args.append(transformed)

    async def _parse_keyword_argument(self, ctx: Context, param: inspect.Parameter, kwargs: dict[str, Any]) -> None:
        kwargs[param.name] = await (  # kwarg only param denotes "consume rest" semantics
            self._transform(ctx, param, ctx.lex.rest) if ctx.lex.rest else self._get_default(ctx, param)
        )

    async def _parse_var_keyword_argument(self, ctx: Context, param: inspect.Parameter, kwargs: dict[str, Any]) -> None:
        kv_pairs = [arg.split("=") for arg in ctx.lex]
        if not kv_pairs:
            raise MissingRequiredArgument(param)  # defaults don't work here

        key_type, value_type = (
            (str, str) if param.annotation in (param.empty, dict) else get_args(param.annotation)
        )  # default to dict[str, str]

        key_converter = self._get_converter(key_type)
        value_converter = self._get_converter(value_type)
        try:
            for key_arg, value_arg in kv_pairs:
                if key_arg in kwargs:
                    raise DuplicateKeywordArgument(key_arg)
                kwargs[await self._convert(ctx, key_converter, param, key_arg.strip())] = await self._convert(
                    ctx, value_converter, param, value_arg.strip()
                )

        except ValueError:
            raise UnmatchedKeyValuePair("Unmatched key-value pair passed") from None

    async def _parse_var_position_argument(self, ctx: Context, param: inspect.Parameter, args: list[Any]) -> None:
        for arg in ctx.lex:
            transformed = await self._transform(ctx, param, arg)
            args.append(transformed)

    async def _parse_arguments(self, ctx: Context) -> None:
        args: list[Any] = []
        kwargs: dict[str, Any] = {}  # these are mutated by functions above

        for param in self.clean_params.values():
            ctx.current_param = param
            if param.kind == param.POSITIONAL_OR_KEYWORD:
                await self._parse_positional_or_keyword_argument(ctx, param, args)
            elif param.kind == param.KEYWORD_ONLY:
                await self._parse_keyword_argument(ctx, param, kwargs)
                break
            elif param.kind == param.VAR_KEYWORD:  # same as **kwargs
                await self._parse_var_keyword_argument(ctx, param, kwargs)
                break
            elif param.kind == param.VAR_POSITIONAL:  # same as *args
                await self._parse_var_position_argument(ctx, param, args)
                break

        ctx.args = tuple(args)
        ctx.kwargs = kwargs

    def _transform(self, ctx: Context, param: inspect.Parameter, argument: str) -> Coroutine[None, None, Any]:
        parm_type = self._prepare_param(param)
        converter = self._get_converter(parm_type)
        return self._convert(ctx, converter, param, argument)

    def _prepare_param(self, param: inspect.Parameter) -> type:
        converter = param.annotation
        if converter is param.empty:
            if param.default is not param.empty:
                converter = str if param.default is None else type(param.default)
            else:
                converter = str
        return converter

    def _get_converter(self, param_type: type) -> converters.Converters:
        converters_ = converters.CONVERTERS.get(param_type, param_type)
        if isinstance(converters_, list):
            if len(converters_) == 1 or not self.special_converters:
                return converters_[0]
            for converter in converters_:
                try:
                    idx = self.special_converters.index(converter)
                except ValueError:
                    pass
                else:
                    return self.special_converters[idx]
            return converters_[0]
        return converters_

    async def _convert(
        self,
        ctx: Context,
        converter: converters.Converters,
        param: inspect.Parameter,
        argument: str,
    ) -> Any:
        if isinstance(converter, type) and issubclass(converter, converters.ConverterBase):
            converter = converter()
        if isinstance(converter, converters.ConverterBase):
            try:
                return await converter.convert(ctx, argument)
            except Exception as exc:
                raise BadArgument(f"{argument!r} failed to convert to {converter.__class__.__name__}") from exc
        origin = get_origin(converter)
        if origin is not None:
            args = get_args(converter)
            for arg in args:
                converter = self._get_converter(arg)
                try:
                    ret = await self._convert(ctx, converter, param, argument)
                except BadArgument:
                    if origin not in (Union, UnionType):
                        raise
                else:
                    if origin is not Literal:
                        return ret
                    if arg == ret:
                        return ret

            if origin in (Union, UnionType) and args[-1] is type(None):  # typing.Optional
                try:
                    return self._get_default(ctx, param)  # get the default if possible
                except MissingRequiredArgument:
                    return None  # fall back to None
            if origin is Literal:
                raise BadArgument(f"Expected one of {', '.join(args)} not {argument!r}")

            raise BadArgument(f"Failed to parse {argument!r} to any type")

        try:
            return converter(argument)
        except Exception as exc:
            try:
                name = converter.__name__  # type: ignore
            except AttributeError:
                name = converter.__class__.__name__
            raise BadArgument(f"{argument!r} failed to convert to {name}") from exc

    async def _get_default(self, ctx: Context, param: inspect.Parameter) -> Any:
        default = param.default
        if default is param.empty:
            raise MissingRequiredArgument(param)
        if isinstance(default, type) and issubclass(default, converters.ConverterBase):
            default = default()
        if isinstance(default, converters.Default):
            try:
                return await default.default(ctx)
            except Exception as exc:
                raise BadArgument(f"{param.default.__class__.__name__} failed to return a default argument") from exc
        return default


CallbackT: TypeAlias = CoroFunc[Concatenate[CogT, Context[Any], P], R] | CoroFunc[Concatenate[Context[Any], P], R]


class GroupMixin(Generic[CogT]):
    """Mixin for something that can have commands registered under it.

    Attributes
    ----------
    case_insensitive
        Whether commands should be invoke-able case insensitively.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        self.case_insensitive: bool = kwargs.get("case_insensitive", False)
        self.__commands__: dict[str, Command[CogT]] = CaseInsensitiveDict() if self.case_insensitive else {}
        super().__init__(*args, **kwargs)

    @property
    def all_commands(self) -> list[Command[CogT]]:
        """A list of the loaded commands."""
        return list(self.__commands__.values())

    @property
    def commands(self) -> set[Command[CogT]]:
        """A set of the loaded commands without duplicates."""
        return set(self.__commands__.values())

    @property
    def children(self) -> Sequence[Command[CogT]]:
        """The commands children."""
        commands = list[Command[CogT]]()
        for command in self.commands:
            commands.append(command)
            if isinstance(command, Group):
                commands += command.children

        return commands

    def add_command(self, command: Command[CogT]) -> None:
        """Add a command to the internal commands list.

        Parameters
        ----------
        command
            The command to register.
        """
        if not isinstance(command, Command):
            raise TypeError("command should derive from commands.Command")

        if command.name in self.__commands__:
            raise ValueError(f"The command {command.name} is already registered.")

        for param in command.clean_params.values():
            if isinstance(param.annotation, str):
                raise TypeError(f"Please rename the parameter {param.name} or make its annotation defined at runtime")

        self.__commands__[command.name] = command
        for alias in command.aliases:
            if alias in self.__commands__:
                self.remove_command(command.name)
                raise ValueError(f"{alias} is already an existing command or alias.")
            self.__commands__[alias] = command

    def remove_command(self, name: str) -> Command[CogT] | None:
        """Remove a command from the internal commands list.

        Parameters
        ----------
        name
            The name of the command to remove.

        Returns
        --------
        The removed command or ``None`` if the command was not found.
        """
        try:
            command = self.__commands__.pop(name)
        except ValueError:
            return None

        for alias in command.aliases:
            del self.__commands__[alias]
        return command

    def get_command(self, name: str) -> Command[CogT] | None:
        """Get a command.

        Parameters
        ----------
        name
            The name of the command.

        Returns
        -------
        The found command or ``None``.
        """

        # fast path, no space in name
        if " " not in name:
            return self.__commands__.get(name)

        names = name.split()
        if not names:
            return None
        command = self.__commands__.get(names[0])
        if not isinstance(command, Group):
            return command

        for name in names[1:]:
            try:
                command = command.__commands__[name]  # type: ignore
            except (AttributeError, KeyError):
                return None

        return command

    @overload
    def command(self, callback: CallbackT[CogT, P, R], /) -> Command[CogT, P, R]: ...

    @overload  # this needs higher kinded types as cls should be type[C[P]] | None
    def command(self, **kwargs: Unpack[CommandKwargsNoCls]) -> Callable[
        [CallbackT[CogT, P, R]],
        Command[CogT, P, R],
    ]: ...

    @overload  # this also needs higher kinded types as cls should be type[C[P]] | None
    def command(self, **kwargs: Unpack[CommandKwargs[C]]) -> Callable[[CallbackT[CogT, P, R]], C]: ...

    def command(  # type: ignore
        self,
        callback: CoroFuncT | None = None,
        /,
        *,
        name: str | None = None,
        cls: type[C] = Command,
        **attrs: Any,
    ) -> Callable[[CoroFuncT], C] | C:
        """|maybecallabledeco|
        A decorator that invokes :func:`command` and adds the created :class:`Command` to the internal command list.

        Parameters
        ----------
        name
            The name of the command. Will default to ``callback.__name__``.
        cls: type[:class:`Command`]
            The class to construct the command from. Defaults to :class:`Command`.
        attrs
            The parameters to pass to the command's ``__init__``.

        Returns
        -------
        The created command.
        """

        def decorator(callback: CoroFuncT) -> C:
            attrs.setdefault("parent", self)
            result = command(name=name, cls=cls, **attrs)(callback)
            self.add_command(result)  # type: ignore
            return result

        return decorator(callback) if callback is not None else decorator

    @overload
    def group(self, callback: CallbackT[CogT, P, R], /) -> Group[CogT, P, R]: ...

    @overload
    def group(self, **kwargs: Unpack[CommandKwargsNoCls]) -> Callable[[CallbackT[CogT, P, R]], Group[CogT, P, R]]: ...

    @overload
    def group(self, **kwargs: Unpack[CommandKwargs[G]]) -> Callable[[CallbackT[CogT, P, R]], G]: ...

    def group(  # type: ignore
        self,
        callback: CoroFuncT | None = None,
        /,
        *,
        name: str | None = None,
        cls: type[G] | None = None,
        **attrs: Any,
    ) -> Callable[[CoroFuncT], G] | G:
        """A decorator that invokes :func:`group` and adds the created :class:`Group` to the internal command list.

        Parameters
        ----------
        name
            The name of the command. Will default to ``callback.__name__``.
        cls: type[:class:`Group`]
            The class to construct the command from. Defaults to :class:`Group`.
        attrs
            The parameters to pass to the command's ``__init__``.

        Returns
        -------
        The created group command.
        """

        def decorator(callback: CoroFuncT) -> G:
            attrs.setdefault("parent", self)
            result = group(name=name, cls=cls or Group, **attrs)(callback)
            self.add_command(result)
            return cast(G, result)  # casting shouldn't really be necessary

        return decorator(callback) if callback is not None else decorator

    def remove_all_commands(self) -> None:
        for command in self.commands:
            if isinstance(command, GroupMixin):
                command.remove_all_commands()
            self.remove_command(command.name)


class Group(GroupMixin[CogT], Command[CogT, P, R]):
    pass


@overload
def command(callback: CallbackT[CogT, P, R], /) -> Command[CogT, P, R]: ...


@overload
def command(**kwargs: Unpack[CommandKwargsNoCls]) -> Callable[[CallbackT[CogT, P, R]], Command[CogT, P, R]]: ...


@overload
def command(**kwargs: Unpack[CommandKwargs[C]]) -> Callable[[CallbackT[CogT, P, R]], C]: ...


def command(  # type: ignore
    callback: CoroFuncT | None = None,
    /,
    *,
    name: str | None = None,
    cls: type[C] = Command,
    **attrs: Any,
) -> Callable[[CoroFuncT], C] | C:
    """|maybecallabledeco|
    A decorator that turns a :term:`coroutine function` into a :class:`Command`.

    Parameters
    ----------
    name
        The name of the command. Will default to ``callback.__name__``.
    cls: type[:class:`Command`]
        The class to construct the command from. Defaults to :class:`Command`.
    attrs
        The attributes to pass to the command's ``__init__``.

    Returns
    -------
    The created command.
    """
    cls = cls or Command

    def decorator(callback: CoroFuncT) -> C:
        if isinstance(callback, Command):
            raise TypeError("callback is already a command.")
        return cls(callback, name=name, **attrs)

    return decorator(callback) if callback is not None else decorator


G = TypeVar("G", bound="Group[Any]", default="Group[Any]")


@overload
def group(callback: CallbackT[CogT, P, R], /) -> Group[CogT, P, R]: ...


@overload
def group(**kwargs: Unpack[CommandKwargsNoCls]) -> Callable[[CallbackT[CogT, P, R]], Group[CogT, P, R]]: ...


@overload
def group(**kwargs: Unpack[CommandKwargs[G]]) -> Callable[[CallbackT[CogT, P, R]], G]: ...


def group(
    callback: CoroFuncT | None = None,
    /,
    *,
    name: str | None = None,
    cls: type[G] | None = None,
    **attrs: Any,
) -> Callable[[CoroFuncT], G] | G:
    """|maybecallabledeco|
    A decorator that turns a :term:`coroutine function` into a :class:`Group`.

    Parameters
    ----------
    name
        The name of the command. Will default to ``callback.__name__``.
    cls: type[:class:`Group`]
        The class to construct the command from. Defaults to :class:`Group`.
    attrs
        The attributes to pass to the command's ``__init__``.

    Returns
    -------
    The created group command.
    """
    return command(callback, name=name, cls=cls or Group, **attrs)  # type: ignore


def check(predicate: Callable[[Context], MaybeBool]) -> Check[MaybeBool]:
    """
    A decorator that registers a function that *could be a* |coroutine_link|_ as a check to a command.

    They should take a singular argument representing the :class:`~steam.ext.commands.Context` for the message.

    Examples
    --------
    .. code:: python

        @commands.check
        def is_mod(ctx: commands.Context) -> bool:
            return ctx.clan and ctx.author in ctx.clan.mods


        @is_mod
        @bot.command
        async def kick(ctx: commands.Context, user: steam.User) -> None: ...

    This will raise an :exc:`steam.ext.commands.CheckFailure` if the user is not an a mod in the clan.

    Attributes
    ----------
    predicate
        The registered check, this will always be a :term:`coroutine function` even if the original check wasn't.
    """

    @overload
    def inner(func: Context, /) -> MaybeBool: ...

    @overload
    def inner(func: MaybeCommandT, /) -> MaybeCommandT: ...

    @functools.partial(cast, Check[MaybeBool])
    def inner(func: Context | MaybeCommandT, /) -> MaybeBool | MaybeCommandT:
        if isinstance(func, Context):
            return predicate(func)
        elif isinstance(func, Command):
            func.checks.append(inner)
        else:
            if not hasattr(func, "__commands_checks__"):
                func.__commands_checks__ = [inner]  # type: ignore
            else:
                func.__commands_checks__.append(inner)  # type: ignore

        return func

    if inspect.iscoroutinefunction(predicate):
        inner.predicate = predicate
    else:

        @functools.wraps(predicate)
        async def wrapper(ctx: Context) -> bool:
            return predicate(ctx)  # type: ignore

        inner.predicate = wrapper

    return inner


@check
def is_owner(ctx: Context) -> bool:
    """
    A decorator that will only allow the bot's owner(s) to invoke the command.

    Warning
    -------
    This relies on :attr:`~steam.ext.commands.Bot.owner_id` or :attr:`~steam.ext.commands.Bot.owner_ids` to be set to
    function if they are not no one will be able to use these commands.
    """

    if ctx.author.id64 == ctx.bot.owner_id or ctx.author.id64 in ctx.bot.owner_ids:
        return True
    raise NotOwner()


@check
def chat_only(ctx: Context) -> bool:
    """A decorator that will make a command only invokable in a :class:`steam.Chat`."""

    if isinstance(ctx.channel, Chat):
        return True

    raise ChatOnly()


@check
def dm_only(ctx: Context) -> bool:
    """A decorator that will make a command only invokable in a :class:`steam.UserChannel`."""

    if isinstance(ctx.channel, UserChannel):
        return True

    raise UserChannelOnly()


def cooldown(rate: int, per: float, type: BucketType = BucketType.Default) -> CommandDeco:
    """Give a :class:`Command`'s a cooldown.

    Parameters
    ----------
    rate
        The amount of times a command can be executed
        before being put on cooldown.
    per
        The amount of time to wait between cooldowns.
    type
        The bucket that the cooldown applies to.

    Examples
    --------
    Usage

    .. code:: python

        @bot.command
        @commands.cooldown(rate=1, per=10, type=commands.BucketType.User)
        async def once_every_ten_seconds(
            ctx: commands.Context,
        ) -> None: ...  # this can only be invoked a user every ten seconds.
    """

    def decorator(command: MaybeCommandT) -> MaybeCommandT:
        if isinstance(command, Command):
            command.cooldown.append(Cooldown(rate, per, type))
        else:
            if not hasattr(command, "__commands_cooldown__"):
                command.__commands_cooldown__ = []  # type: ignore
            command.__commands_cooldown__.append(Cooldown(rate, per, type))  # type: ignore
        return command

    return decorator



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, Literal, cast, overload

from typing_extensions import TypeVar

from . import utils
from ._const import URL
from .enums import LeaderboardDataRequest, LeaderboardDisplayType, LeaderboardSortMethod, LeaderboardUploadScoreMethod
from .user import User, WrapsUser

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Sequence

    from .app import App
    from .state import ConnectionState
    from .types.id import ID64, LeaderboardID
    from .types.user import IndividualID


__all__ = (
    "LeaderboardScoreUpdate",
    "LeaderboardUser",
    "Leaderboard",
)


@dataclass(slots=True)
class LeaderboardScoreUpdate:
    """Represents a score update on a leaderboard."""

    global_rank_before: int
    """The global rank of the user before the score update."""
    global_rank_after: int
    """The global rank of the user after the score update."""
    score_changed: bool
    """Whether the score changed."""


class LeaderboardUser(WrapsUser):
    """Represents a user on an app's leaderboard."""

    __slots__ = ("global_rank", "score", "details", "ugc_id")

    def __init__(
        self,
        state: ConnectionState,
        user: User,
        global_rank: int,
        score: int,
        details: bytes,
        ugc_id: int,
    ):
        super().__init__(state, user)
        self.global_rank = global_rank
        """The global rank of the user."""
        self.score = score
        """The score of the user."""
        self.details = details
        """The details of the user."""
        self.ugc_id = ugc_id
        """The UGC ID of the user."""


AppT = TypeVar("AppT", bound="App", covariant=True)
DisplayNameT = TypeVar("DisplayNameT", bound=str | None, default=None, covariant=True)


@dataclass(slots=True)
class Leaderboard(Generic[AppT, DisplayNameT]):
    """Represents a leaderboard for an app."""

    _state: ConnectionState
    id: LeaderboardID
    """The leaderboard's id."""
    name: str
    """The name of the leaderboard."""
    app: AppT
    """The app this leaderboard is for."""
    sort_method: LeaderboardSortMethod
    """The sort method of the leaderboard."""
    display_type: LeaderboardDisplayType
    """The display type of the leaderboard."""
    entry_count: int
    """The number of entries in the leaderboard."""
    display_name: DisplayNameT = cast(DisplayNameT, None)
    """The display name of the leaderboard. This is only set if the leaderboard is fetched using :meth:`PartialApp.leaderboards`."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id} display_name={self.display_name!r} app={self.app!r}>"

    @property
    def url(self) -> str:
        """The URL of the leaderboard."""
        return f"{URL.COMMUNITY}/stats/{self.app.id}/leaderboards/{self.id}"

    @overload
    async def entries(
        self,
        *,
        limit: int | None = None,
        type: Literal[LeaderboardDataRequest.Users],
        users: Sequence[IndividualID],
    ) -> AsyncGenerator[LeaderboardUser, None]: ...

    @overload
    async def entries(
        self, *, limit: int | None = None, type: LeaderboardDataRequest = LeaderboardDataRequest.Global
    ) -> AsyncGenerator[LeaderboardUser, None]: ...

    async def entries(
        self,
        *,
        limit: int | None = None,
        type: LeaderboardDataRequest = LeaderboardDataRequest.Global,
        users: Sequence[IndividualID] = (),
    ) -> AsyncGenerator[LeaderboardUser, None]:
        """Fetch the entries of the leaderboard.

        Parameters
        ----------
        limit
            The maximum number of entries to fetch. If `None` is passed, all entries will be fetched.
        type
            The type of entries to fetch.
        users
            The users to fetch. Only used if `type` is :attr:`LeaderboardDataRequest.Users`.
        """
        id64s = [user.id64 for user in users]
        for start, stop in utils._int_chunks(min(self.entry_count, limit or self.entry_count), 100):
            entries = await self._state.fetch_app_leaderboard_entries(self.app.id, self.id, start, stop, type, id64s)
            for entry, user in zip(
                entries,
                await self._state._maybe_users(cast("list[ID64]", [entry.steam_id_user for entry in entries])),
            ):
                yield LeaderboardUser(
                    self._state,
                    user,
                    entry.global_rank,
                    entry.score,
                    entry.details,
                    entry.ugc_id,
                )

    async def set_score(
        self, score: int, *, details: bytes, method: LeaderboardUploadScoreMethod = LeaderboardUploadScoreMethod.NONE
    ) -> LeaderboardScoreUpdate:
        """Set the score of the current user on the leaderboard.

        Parameters
        ----------
        score
            The score to set.
        details
            The details to set.
        method
            The method to use to upload the score.
        """
        update = await self._state.set_app_leaderboard_score(self.app.id, self.id, score, details, method)
        return LeaderboardScoreUpdate(update.global_rank_previous, update.global_rank_new, update.score_changed)

    async def set_ugc_id(self, ugc_id: int) -> None:
        """Set the UGC ID of the current user on the leaderboard."""
        await self._state.set_app_leaderboard_ugc(self.app.id, self.id, ugc_id)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: tf_gcmessages.proto
# plugin: python-betterproto
# Last updated

from dataclasses import dataclass
from typing import TYPE_CHECKING

import betterproto

if TYPE_CHECKING:
    from .base import Item


class EtfMatchGroup(betterproto.Enum):
    Invalid = -1
    MvMPractice = 0
    MvMMannUp = 1
    First = 0
    MvMDefault = 0
    MvMFirst = 0
    MvMLast = 1
    Ladder6v6 = 2
    Ladder9v9 = 3
    Ladder12v12 = 4
    LadderDefault = 2
    LadderFirst = 2
    LadderLast = 4
    Casual6v6 = 5
    Casual9v9 = 6
    Casual12v12 = 7
    CasualDefault = 7
    CasualFirst = 5
    CasualLast = 7
    EventPlaceholder = 8
    EventDefault = 8
    EventFirst = 8
    EventLast = 8


class ETFPartyChatType(betterproto.Enum):
    Invalid = 0
    MemberChat = 1
    SyntheticMemberJoin = 1000
    SyntheticMemberLeave = 1001
    SyntheticSendFailed = 1002
    SyntheticMemberOnline = 1003
    SyntheticMemberOffline = 1004


@dataclass(eq=False, repr=False)
class SaxxyBroadcast(betterproto.Message):
    category_number: int = betterproto.int32_field(1)
    user_name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class SpecificItemBroadcast(betterproto.Message):
    item_def_index: int = betterproto.uint32_field(1)
    was_destruction: bool = betterproto.bool_field(2)
    user_name: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class WorldStatus(betterproto.Message):
    beta_stress_test_event_active: bool = betterproto.bool_field(1)
    event_match_group: "EtfMatchGroup" = betterproto.enum_field(2)
    event_expire_time: int = betterproto.fixed32_field(3)
    active_client_version: int = betterproto.uint32_field(4)
    disabled_match_groups: "list[EtfMatchGroup]" = betterproto.enum_field(5)


@dataclass(eq=False, repr=False)
class DuelSummary(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    duel_wins: int = betterproto.uint32_field(2)
    duel_losses: int = betterproto.uint32_field(3)
    last_duel_account_id: int = betterproto.uint32_field(4)
    last_duel_timestamp: int = betterproto.uint32_field(5)
    last_duel_status: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class MapContribution(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    def_index: int = betterproto.uint32_field(2)
    contribution_level: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class FreeTrialChooseMostHelpfulFriend(betterproto.Message):
    account_id_friend: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class RequestTF2Friends(betterproto.Message):
    account_ids: list[int] = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class RequestTF2FriendsResponse(betterproto.Message):
    account_ids: list[int] = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class PlayerInfo(betterproto.Message):
    num_new_users_helped: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ThankedBySomeone(betterproto.Message):
    thanker_steam_id: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class ThankedSomeone(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class FreeTrialConvertedToPremium(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class SaxxyAwarded(betterproto.Message):
    category: int = betterproto.uint32_field(1)
    winner_names: list[str] = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ReplaySubmitContestEntry(betterproto.Message):
    youtube_url: str = betterproto.string_field(1)
    category: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class ReplaySubmitContestEntryResponse(betterproto.Message):
    success: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class ReplayCachedContestData(betterproto.Message):
    timestamp: int = betterproto.fixed32_field(1)
    num_votes_last_day: int = betterproto.uint32_field(2)
    video_entry_ids: list[int] = betterproto.uint32_field(3)
    num_flags_last_day: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class CoachingAddToCoaches(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CoachingRemoveFromCoaches(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class CoachingFindCoach(betterproto.Message):
    account_id_friend_as_coach: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CoachingFindCoachResponse(betterproto.Message):
    found_coach: bool = betterproto.bool_field(1)
    num_likes: int = betterproto.uint32_field(2)
    coach_name: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CoachingAskCoach(betterproto.Message):
    account_id_student: int = betterproto.uint32_field(1)
    student_is_friend: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class CoachingAskCoachResponse(betterproto.Message):
    accept_coaching_assignment: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class CoachingCoachJoinGame(betterproto.Message):
    join_game: bool = betterproto.bool_field(1)
    server_address: int = betterproto.uint32_field(2)
    server_port: int = betterproto.uint32_field(3)
    account_id_student: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class CoachingCoachJoining(betterproto.Message):
    account_id_coach: int = betterproto.uint32_field(1)
    account_id_student: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CoachingCoachJoined(betterproto.Message):
    account_id_coach: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CoachingLikeCurrentCoach(betterproto.Message):
    like_coach: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class CoachingRemoveCurrentCoach(betterproto.Message):
    account_id_coach: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class QuickplayScoreServers(betterproto.Message):
    servers: "list[QuickplayScoreServersServerInfo]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuickplayScoreServersServerInfo(betterproto.Message):
    server_address: int = betterproto.uint32_field(1)
    server_port: int = betterproto.uint32_field(2)
    num_users: int = betterproto.uint32_field(3)
    steam_id: int = betterproto.uint64_field(4)
    max_users: int = betterproto.uint32_field(5)
    user_score: float = betterproto.float_field(6)


@dataclass(eq=False, repr=False)
class QuickplayScoreServersResponse(betterproto.Message):
    servers: "list[QuickplayScoreServersResponseServerInfo]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class QuickplayScoreServersResponseServerInfo(betterproto.Message):
    server_address: int = betterproto.uint32_field(1)
    server_port: int = betterproto.uint32_field(2)
    total_score: float = betterproto.float_field(3)
    steam_id: int = betterproto.uint64_field(4)
    options_score: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class QuickplayPlayerJoining(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ServerLevelInfo(betterproto.Message):
    level_loaded: bool = betterproto.bool_field(1)
    level_name: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ServerAuthChallenge(betterproto.Message):
    challenge_string: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ServerAuthResult(betterproto.Message):
    authenticated: bool = betterproto.bool_field(1)
    game_server_standing: int = betterproto.int32_field(2)
    game_server_standing_trend: int = betterproto.int32_field(3)
    is_valve_server: bool = betterproto.bool_field(4)
    message: str = betterproto.string_field(5)


@dataclass(eq=False, repr=False)
class ServerAuthChallengeResponse(betterproto.Message):
    game_server_account_id: int = betterproto.uint32_field(1)
    hashed_challenge_string: bytes = betterproto.bytes_field(2)


@dataclass(eq=False, repr=False)
class ServerCreateIdentity(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ServerCreateIdentityResponse(betterproto.Message):
    account_created: bool = betterproto.bool_field(1)
    game_server_account_id: int = betterproto.uint32_field(2)
    game_server_identity_token: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class ServerList(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ServerListResponse(betterproto.Message):
    owned_game_servers: "list[ServerListResponseGameServerIdentity]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ServerListResponseGameServerIdentity(betterproto.Message):
    game_server_account_id: int = betterproto.uint32_field(1)
    game_server_identity_token: str = betterproto.string_field(2)
    game_server_standing: int = betterproto.int32_field(3)
    game_server_standing_trend: int = betterproto.int32_field(4)


@dataclass(eq=False, repr=False)
class ServerResetIdentity(betterproto.Message):
    game_server_account_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ServerResetIdentityResponse(betterproto.Message):
    game_server_identity_token_reset: bool = betterproto.bool_field(1)
    game_server_account_id: int = betterproto.uint32_field(2)
    game_server_identity_token: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class ServerAckPolicy(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ServerAckPolicyResponse(betterproto.Message):
    result: int = betterproto.uint32_field(1)
    message: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ClientUseServerModificationItem(betterproto.Message):
    item_id: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class HalloweenReservedItem(betterproto.Message):
    x: list[float] = betterproto.float_field(1)
    y: list[float] = betterproto.float_field(2)
    z: list[float] = betterproto.float_field(3)
    spawn_meta_info: int = betterproto.uint32_field(7)


@dataclass(eq=False, repr=False)
class HalloweenGrantItem(betterproto.Message):
    recipient_account_id: int = betterproto.uint32_field(1)
    level_id: int = betterproto.uint32_field(2)
    flagged: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class HalloweenGrantItemResponse(betterproto.Message):
    recipient_account_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class HalloweenItemClaimed(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PickupItemEligibilityQuery(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    seconds_ago: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class PickupItemEligibilityQueryResponse(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    was_eligible: bool = betterproto.bool_field(2)
    level_id: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class PartyMember(betterproto.Message):
    owns_ticket: bool = betterproto.bool_field(2)
    completed_missions: int = betterproto.uint32_field(3)
    badge_level: int = betterproto.uint32_field(4)
    competitive_access: bool = betterproto.bool_field(9)
    experience: int = betterproto.uint32_field(14)
    player_criteria: "PerPlayerMatchCriteriaProto" = betterproto.message_field(16)
    activity: "PartyMemberActivity" = betterproto.message_field(17)
    casual_banned: bool = betterproto.bool_field(18)
    ranked_banned: bool = betterproto.bool_field(19)
    casual_low_priority: bool = betterproto.bool_field(20)
    ranked_low_priority: bool = betterproto.bool_field(21)
    lobby_standby: bool = betterproto.bool_field(22)


@dataclass(eq=False, repr=False)
class PartyMemberActivity(betterproto.Message):
    lobby_id: int = betterproto.fixed64_field(1)
    lobby_match_group: "EtfMatchGroup" = betterproto.enum_field(2)
    multiqueue_blocked: bool = betterproto.bool_field(3)
    online: bool = betterproto.bool_field(4)
    client_version: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class TFPendingPartyMember(betterproto.Message):
    steamid: int = betterproto.fixed64_field(1)
    inviter: int = betterproto.fixed64_field(3)


@dataclass(eq=False, repr=False)
class GroupMatchCriteriaProto(betterproto.Message):
    late_join_ok: bool = betterproto.bool_field(5)
    custom_ping_tolerance: int = betterproto.uint32_field(13)
    mvm_mannup_tour: str = betterproto.string_field(10)
    mvm_mannup_missions: list[str] = betterproto.string_field(15)
    mvm_bootcamp_missions: list[str] = betterproto.string_field(16)
    casual_criteria: "CasualMatchCriteria" = betterproto.message_field(12)


@dataclass(eq=False, repr=False)
class CasualMatchCriteria(betterproto.Message):
    selected_maps_bits: list[int] = betterproto.fixed32_field(3)


@dataclass(eq=False, repr=False)
class PerPlayerMatchCriteriaProto(betterproto.Message):
    mvm_squad_surplus: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class PartyOptions(betterproto.Message):
    overwrite_existing: bool = betterproto.bool_field(1)
    group_criteria: "GroupMatchCriteriaProto" = betterproto.message_field(2)
    player_criteria: "PerPlayerMatchCriteriaProto" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class PartySetOptions(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)
    options: "PartyOptions" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class PartySetOptionsResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PartyQueueForMatch(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)
    final_options: "PartyOptions" = betterproto.message_field(2)
    match_group: "EtfMatchGroup" = betterproto.enum_field(3)


@dataclass(eq=False, repr=False)
class PartyQueueForMatchResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PartyQueueForStandby(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)
    party_lobby_id: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class PartyQueueForStandbyResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PartyRemoveFromQueue(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)
    match_group: "EtfMatchGroup" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class PartyRemoveFromQueueResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PartyRemoveFromStandbyQueue(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class PartyRemoveFromStandbyQueueResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PartyInvitePlayer(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)
    player_id: int = betterproto.fixed64_field(2)
    expecting_request_to_join: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class PartyRequestJoinPlayer(betterproto.Message):
    current_party_id: int = betterproto.fixed64_field(1)
    join_player_id: int = betterproto.fixed64_field(2)
    join_party_id: int = betterproto.fixed64_field(3)
    expecting_invite: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class PartyClearPendingPlayer(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)
    pending_player_id: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class PartyClearPendingPlayerResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PartyClearOtherPartyRequest(betterproto.Message):
    other_party_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class PartyClearOtherPartyRequestResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PartyPromoteToLeader(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)
    new_leader_id: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class PartyKickMember(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)
    target_id: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class PartySendChat(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)
    msg: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class PartyChatMsg(betterproto.Message):
    type: "ETFPartyChatType" = betterproto.enum_field(1)
    actor_id: int = betterproto.fixed64_field(2)
    msg: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class Party(betterproto.Message):
    party_id: int = betterproto.uint64_field(1)
    leader_id: int = betterproto.fixed64_field(2)
    member_ids: list[int] = betterproto.fixed64_field(3)
    members: "list[PartyMember]" = betterproto.message_field(13)
    associated_lobby_id: int = betterproto.uint64_field(35)
    associated_lobby_match_group: "EtfMatchGroup" = betterproto.enum_field(40)
    matchmaking_queues: "list[PartyQueueEntry]" = betterproto.message_field(43)
    group_criteria: "GroupMatchCriteriaProto" = betterproto.message_field(37)
    casual_banned_time: int = betterproto.uint32_field(18)
    casual_low_priority_time: int = betterproto.uint32_field(20)
    ranked_banned_time: int = betterproto.uint32_field(41)
    ranked_low_priority_time: int = betterproto.uint32_field(42)
    pending_members: "list[TFPendingPartyMember]" = betterproto.message_field(39)


@dataclass(eq=False, repr=False)
class PartyQueueEntry(betterproto.Message):
    match_group: "EtfMatchGroup" = betterproto.enum_field(1)
    queued_time: int = betterproto.fixed32_field(2)


@dataclass(eq=False, repr=False)
class PartyInvite(betterproto.Message):
    group_id: int = betterproto.uint64_field(1)
    inviter: int = betterproto.fixed64_field(2)
    members: "list[PartyInvitePartyMember]" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class PartyInvitePartyMember(betterproto.Message):
    steamid: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class LobbyPlayerProto(betterproto.Message):
    id: int = betterproto.fixed64_field(1)
    name: str = betterproto.string_field(6)
    original_party_id: int = betterproto.uint64_field(12)
    squad_surplus: bool = betterproto.bool_field(14)
    badge_level: int = betterproto.uint32_field(15)
    last_connect_time: int = betterproto.uint32_field(17)
    normalized_rating: float = betterproto.double_field(20)
    normalized_uncertainty: float = betterproto.double_field(22)
    rank: int = betterproto.uint32_field(21)
    chat_suspension: bool = betterproto.bool_field(23)


@dataclass(eq=False, repr=False)
class LobbyInviteProto(betterproto.Message):
    lobby_id: int = betterproto.fixed64_field(1)
    match_group: "EtfMatchGroup" = betterproto.enum_field(2)


@dataclass(eq=False, repr=False)
class GameServerLobby(betterproto.Message):
    lobby_id: int = betterproto.uint64_field(1)
    members: "list[LobbyPlayerProto]" = betterproto.message_field(2)
    server_id: int = betterproto.fixed64_field(6)
    connect: str = betterproto.string_field(5)
    initial_average_mm_rating: float = betterproto.double_field(32)
    mannup_tour_name: str = betterproto.string_field(42)
    map_name: str = betterproto.string_field(38)
    mission_name: str = betterproto.string_field(39)
    match_group: int = betterproto.uint32_field(41)
    match_id: int = betterproto.uint64_field(30)
    formed_time: int = betterproto.uint32_field(36)
    flags: int = betterproto.uint32_field(43)
    late_join_eligible: bool = betterproto.bool_field(44)
    fixed_match_size: int = betterproto.uint32_field(45)
    next_maps_for_vote: list[int] = betterproto.uint32_field(47)
    lobby_mm_version: int = betterproto.uint32_field(48)
    pending_members: "list[LobbyPlayerProto]" = betterproto.message_field(49)


@dataclass(eq=False, repr=False)
class ExitMatchmaking(betterproto.Message):
    explicit_abandon: bool = betterproto.bool_field(1)
    party_id: int = betterproto.uint64_field(2)
    lobby_id: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class AcceptLobbyInvite(betterproto.Message):
    invited_lobby_id: int = betterproto.uint64_field(1)
    abandoning_match_id: int = betterproto.uint64_field(2)
    abandoning_invite_lobby_ids: list[int] = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class AcceptLobbyInviteReply(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class MatchmakingSearchCountRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class MatchmakingSearchCountResponse(betterproto.Message):
    searching_players_by_group: list[int] = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class KickedFromMatchmakingQueue(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GameServerMatchmakingStatus(betterproto.Message):
    server_version: int = betterproto.uint32_field(16)
    map: str = betterproto.string_field(3)
    tags: str = betterproto.string_field(4)
    bot_count: int = betterproto.uint32_field(5)
    num_spectators: int = betterproto.uint32_field(6)
    max_players: int = betterproto.uint32_field(7)
    slots_free: int = betterproto.uint32_field(8)
    server_region: int = betterproto.uint32_field(9)
    server_loadavg: float = betterproto.float_field(10)
    server_trusted: bool = betterproto.bool_field(11)
    server_dedicated: bool = betterproto.bool_field(12)
    strict: int = betterproto.uint32_field(17)
    players: "list[GameServerMatchmakingStatusPlayer]" = betterproto.message_field(13)
    mvm_wave: int = betterproto.uint32_field(18)
    mvm_credits_acquired: int = betterproto.uint32_field(19)
    mvm_credits_dropped: int = betterproto.uint32_field(20)
    match_group: int = betterproto.sint32_field(23)
    lobby_mm_version: int = betterproto.uint32_field(24)


@dataclass(eq=False, repr=False)
class GameServerMatchmakingStatusPlayer(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class MatchmakingProgress(betterproto.Message):
    avg_wait_time_new: int = betterproto.uint32_field(4)
    avg_wait_time_join_late: int = betterproto.uint32_field(5)
    your_wait_time: int = betterproto.uint32_field(6)
    matching_worldwide_searching_players: int = betterproto.uint32_field(8)
    matching_near_you_searching_players: int = betterproto.uint32_field(9)
    total_worldwide_searching_players: int = betterproto.uint32_field(13)
    total_near_you_searching_players: int = betterproto.uint32_field(14)
    matching_worldwide_active_players: int = betterproto.uint32_field(15)
    matching_near_you_active_players: int = betterproto.uint32_field(16)
    total_worldwide_active_players: int = betterproto.uint32_field(17)
    total_near_you_active_players: int = betterproto.uint32_field(18)
    matching_worldwide_empty_gameservers: int = betterproto.uint32_field(19)
    matching_near_you_empty_gameservers: int = betterproto.uint32_field(20)
    total_worldwide_empty_gameservers: int = betterproto.uint32_field(21)
    total_near_you_empty_gameservers: int = betterproto.uint32_field(22)
    urgency_pct: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class MvMVictoryInfo(betterproto.Message):
    players: "list[MvMVictoryInfoPlayer]" = betterproto.message_field(1)
    tour_name: str = betterproto.string_field(2)
    mission_name: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class MvMVictoryInfoItem(betterproto.Message):
    item_data: bytes = betterproto.bytes_field(2)
    squad_surplus_claimer_steam_id: int = betterproto.fixed64_field(3)


@dataclass(eq=False, repr=False)
class MvMVictoryInfoPlayer(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)
    badge_granted: bool = betterproto.bool_field(3)
    badge_progress_updated: bool = betterproto.bool_field(4)
    badge_leveled: bool = betterproto.bool_field(5)
    badge_level: int = betterproto.uint32_field(6)
    badge_progress_bits: int = betterproto.uint32_field(7)
    items: "list[MvMVictoryInfoItem]" = betterproto.message_field(8)
    voucher_missing: bool = betterproto.bool_field(9)
    badge_points: int = betterproto.uint32_field(10)


@dataclass(eq=False, repr=False)
class CgHelloResponse(betterproto.Message):
    version_check: int = betterproto.uint32_field(1)
    version_checksum: list[int] = betterproto.uint64_field(2)
    version_verbose: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CgSync(betterproto.Message):
    version_checksum: bytes = betterproto.bytes_field(1)
    version_check: int = betterproto.uint32_field(2)
    version_check_ex: int = betterproto.uint32_field(3)
    version_check_ex2: int = betterproto.uint32_field(4)
    version_checksum_ex: bytes = betterproto.bytes_field(5)


@dataclass(eq=False, repr=False)
class CgSyncEx(betterproto.Message):
    version_checksum: str = betterproto.string_field(1)
    version_checksum_ex: bytes = betterproto.bytes_field(2)
    version_check: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class MvMVictory(betterproto.Message):
    legacy_mission_index: int = betterproto.uint32_field(1)
    tour_name_mannup: str = betterproto.string_field(5)
    mission_name: str = betterproto.string_field(6)
    players: "list[MvMVictoryPlayer]" = betterproto.message_field(2)
    lobby_id: int = betterproto.uint64_field(3)
    event_time: int = betterproto.fixed32_field(4)


@dataclass(eq=False, repr=False)
class MvMVictoryPlayer(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)
    squad_surplus: bool = betterproto.bool_field(2)


@dataclass(eq=False, repr=False)
class MvMMannUpVictoryReply(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class GameServerKickingLobby(betterproto.Message):
    lobby_id: int = betterproto.uint64_field(3)
    match_id: int = betterproto.uint64_field(4)


@dataclass(eq=False, repr=False)
class GameServerKickingLobbyResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class LeaveGameAndPrepareToJoinParty(betterproto.Message):
    party_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class PlayerLeftMatch(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)
    was_abandon: bool = betterproto.bool_field(3)
    lobby_id: int = betterproto.uint64_field(4)
    match_id: int = betterproto.uint64_field(5)
    xp_breakdown: "list[XPSource]" = betterproto.message_field(6)


@dataclass(eq=False, repr=False)
class PlayerLeftMatchResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PlayerVoteKickedAfterLeavingMatchResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class HalloweenServerBossEvent(betterproto.Message):
    event_counter: int = betterproto.uint32_field(1)
    timestamp: int = betterproto.uint32_field(2)
    boss_type: int = betterproto.uint32_field(3)
    boss_level: int = betterproto.uint32_field(4)
    event_type: int = betterproto.uint32_field(5)
    players_involved: int = betterproto.uint32_field(6)
    elapsed_time: float = betterproto.float_field(7)


@dataclass(eq=False, repr=False)
class CAttributeString(betterproto.Message):
    value: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class CAttributeDynamicRecipeComponent(betterproto.Message):
    def_index: int = betterproto.uint32_field(1)
    item_quality: int = betterproto.uint32_field(2)
    component_flags: int = betterproto.uint32_field(3)
    attributes_string: str = betterproto.string_field(4)
    num_required: int = betterproto.uint32_field(5)
    num_fulfilled: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class CAttributeDynamicRecipeComponentCompatNeverSerializeThisOut(betterproto.Message):
    def_index: int = betterproto.uint32_field(1)
    item_def: int = betterproto.uint32_field(2)
    item_quality: int = betterproto.uint32_field(3)
    component_flags: int = betterproto.uint32_field(4)
    item_flags: int = betterproto.uint32_field(5)
    attributes_string: str = betterproto.string_field(6)
    num_required: int = betterproto.uint32_field(7)
    item_count: int = betterproto.uint32_field(8)
    num_fulfilled: int = betterproto.uint32_field(9)
    items_fulfilled: int = betterproto.uint32_field(10)


@dataclass(eq=False, repr=False)
class CAttributeItemSlotCriteria(betterproto.Message):
    tags: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class SetItemSlotAttribute(betterproto.Message):
    item_id: int = betterproto.uint64_field(1)
    slot_item_original_id: int = betterproto.uint64_field(2)
    slot_index: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CsoWarData(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    war_id: int = betterproto.uint32_field(2)
    affiliation: int = betterproto.uint32_field(3)
    points_scored: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class CgGcWarIndividualUpdate(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)
    war_id: int = betterproto.uint32_field(2)
    score: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class CgGcWarJoinWar(betterproto.Message):
    affiliation: int = betterproto.uint32_field(1)
    war_id: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CgGcWarRequestGlobalStats(betterproto.Message):
    war_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class CgGcWarGlobalStatsResponse(betterproto.Message):
    side_scores: "list[CgGcWarGlobalStatsResponseSideScore]" = betterproto.message_field(1)
    war_id: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class CgGcWarGlobalStatsResponseSideScore(betterproto.Message):
    side: int = betterproto.uint32_field(1)
    score: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CgGcPlayerDuckLeaderboardIndividualUpdate(betterproto.Message):
    score: int = betterproto.uint32_field(2)
    type: int = betterproto.uint32_field(3)
    score_id: bytes = betterproto.bytes_field(4)
    score_check: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class CAttributeWorldItemPlacement(betterproto.Message):
    original_item_id: int = betterproto.uint64_field(1)
    pos_x: float = betterproto.float_field(2)
    pos_y: float = betterproto.float_field(3)
    pos_z: float = betterproto.float_field(4)
    ang_x: float = betterproto.float_field(5)
    ang_y: float = betterproto.float_field(6)
    ang_z: float = betterproto.float_field(7)


@dataclass(eq=False, repr=False)
class CgWorldItemPlacementUpdate(betterproto.Message):
    original_item_id: int = betterproto.uint64_field(1)
    pos_x: float = betterproto.float_field(2)
    pos_y: float = betterproto.float_field(3)
    pos_z: float = betterproto.float_field(4)
    ang_x: float = betterproto.float_field(5)
    ang_y: float = betterproto.float_field(6)
    ang_z: float = betterproto.float_field(7)
    force_remove_all: bool = betterproto.bool_field(8)
    attrib_name: str = betterproto.string_field(9)


@dataclass(eq=False, repr=False)
class AcknowledgeXp(betterproto.Message):
    match_group: int = betterproto.int32_field(1)
    predicted_experience: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class XPSource(betterproto.Message):
    amount: int = betterproto.int32_field(2)
    match_group: int = betterproto.int32_field(3)
    account_id: int = betterproto.uint32_field(4)
    match_id: int = betterproto.uint64_field(5)


@dataclass(eq=False, repr=False)
class XPSourceBreakdown(betterproto.Message):
    sources: "list[XPSource]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ClientInit(betterproto.Message):
    client_version: int = betterproto.uint32_field(1)
    language: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class Notification(betterproto.Message):
    notification_id: int = betterproto.uint64_field(1)
    account_id: int = betterproto.uint32_field(2)
    expiration_time: int = betterproto.fixed32_field(3)
    notification_string: str = betterproto.string_field(5)


@dataclass(eq=False, repr=False)
class NotificationQueue(betterproto.Message):
    notifications: "list[Notification]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class NotificationAcknowledge(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    notification_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class NotificationAcknowledgeReply(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class MatchResult(betterproto.Message):
    match_id: int = betterproto.uint64_field(1)
    match_group: int = betterproto.int32_field(2)
    duration: int = betterproto.uint32_field(4)
    red_score: int = betterproto.uint32_field(5)
    blue_score: int = betterproto.uint32_field(6)
    winning_team: int = betterproto.uint32_field(7)
    map_index: int = betterproto.uint32_field(8)
    game_type: int = betterproto.uint32_field(9)
    players: "list[MatchResultPlayer]" = betterproto.message_field(10)
    win_reason: int = betterproto.uint32_field(13)
    flags: int = betterproto.uint32_field(14)
    bots: int = betterproto.uint32_field(16)


@dataclass(eq=False, repr=False)
class MatchResultPlayer(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)
    original_party_id: int = betterproto.uint64_field(2)
    team: int = betterproto.uint32_field(3)
    score: int = betterproto.uint32_field(4)
    ping: int = betterproto.uint32_field(5)
    flags: int = betterproto.uint32_field(6)
    rank: int = betterproto.uint32_field(9)
    classes_played: int = betterproto.uint32_field(10)
    kills: int = betterproto.uint32_field(11)
    deaths: int = betterproto.uint32_field(12)
    damage: int = betterproto.uint32_field(13)
    healing: int = betterproto.uint32_field(14)
    support: int = betterproto.uint32_field(15)
    score_medal: int = betterproto.uint32_field(16)
    kills_medal: int = betterproto.uint32_field(17)
    damage_medal: int = betterproto.uint32_field(18)
    healing_medal: int = betterproto.uint32_field(19)
    support_medal: int = betterproto.uint32_field(20)
    xp_breakdown: "list[XPSource]" = betterproto.message_field(21)
    leave_time: int = betterproto.uint32_field(22)
    connect_time: int = betterproto.uint32_field(24)


@dataclass(eq=False, repr=False)
class MatchResultResponse(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class PreviewDataBlock(betterproto.Message):
    econitem: "Item" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class PreviewDataBlockRequest(betterproto.Message):
    param_s: int = betterproto.uint64_field(1)
    param_a: int = betterproto.uint64_field(2)
    param_d: int = betterproto.uint64_field(3)
    param_m: int = betterproto.uint64_field(4)


@dataclass(eq=False, repr=False)
class PreviewDataBlockResponse(betterproto.Message):
    iteminfo: "PreviewDataBlock" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class LadderPlayerStats(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    match_group: int = betterproto.int32_field(2)
    season_id: int = betterproto.uint32_field(3)
    games: int = betterproto.uint32_field(9)
    score: int = betterproto.uint32_field(10)
    kills: int = betterproto.uint32_field(11)
    deaths: int = betterproto.uint32_field(12)
    damage: int = betterproto.uint32_field(13)
    healing: int = betterproto.uint32_field(14)
    support: int = betterproto.uint32_field(15)
    score_bronze: int = betterproto.uint32_field(16)
    score_silver: int = betterproto.uint32_field(17)
    score_gold: int = betterproto.uint32_field(18)
    kills_bronze: int = betterproto.uint32_field(19)
    kills_silver: int = betterproto.uint32_field(20)
    kills_gold: int = betterproto.uint32_field(21)
    damage_bronze: int = betterproto.uint32_field(22)
    damage_silver: int = betterproto.uint32_field(23)
    damage_gold: int = betterproto.uint32_field(24)
    healing_bronze: int = betterproto.uint32_field(25)
    healing_silver: int = betterproto.uint32_field(26)
    healing_gold: int = betterproto.uint32_field(27)
    support_bronze: int = betterproto.uint32_field(28)
    support_silver: int = betterproto.uint32_field(29)
    support_gold: int = betterproto.uint32_field(30)


@dataclass(eq=False, repr=False)
class RatingData(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    rating_type: int = betterproto.int32_field(2)
    rating_primary: int = betterproto.uint32_field(3)
    rating_secondary: int = betterproto.uint32_field(4)
    rating_tertiary: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class VoteKickPlayerRequest(betterproto.Message):
    target_id: int = betterproto.uint64_field(2)
    voter_id: int = betterproto.uint64_field(3)
    match_id: int = betterproto.uint64_field(5)


@dataclass(eq=False, repr=False)
class VoteKickPlayerRequestResponse(betterproto.Message):
    allowed: bool = betterproto.bool_field(1)
    voter_inhibit: bool = betterproto.bool_field(3)
    target_inhibit: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class DailyCompetitiveStatsRollup(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DailyCompetitiveStatsRollupResponse(betterproto.Message):
    rankdata: "list[DailyCompetitiveStatsRollupResponseRankBucketEntry]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DailyCompetitiveStatsRollupResponseRankBucketEntry(betterproto.Message):
    rank: int = betterproto.uint32_field(1)
    records: int = betterproto.uint32_field(2)
    avg_score: int = betterproto.uint32_field(3)
    stdev_score: int = betterproto.uint32_field(4)
    avg_kills: int = betterproto.uint32_field(5)
    stdev_kills: int = betterproto.uint32_field(6)
    avg_damage: int = betterproto.uint32_field(7)
    stdev_damage: int = betterproto.uint32_field(8)
    avg_healing: int = betterproto.uint32_field(9)
    stdev_healing: int = betterproto.uint32_field(10)
    avg_support: int = betterproto.uint32_field(11)
    stdev_support: int = betterproto.uint32_field(12)


@dataclass(eq=False, repr=False)
class ReportPlayer(betterproto.Message):
    account_id_target: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class MatchResultPlayerStats(betterproto.Message):
    match_id: int = betterproto.uint64_field(1)
    account_id: int = betterproto.uint32_field(2)
    match_group: int = betterproto.int32_field(3)
    endtime: int = betterproto.uint32_field(4)
    season_id: int = betterproto.uint32_field(5)
    status: int = betterproto.uint32_field(6)
    original_party_id: int = betterproto.uint32_field(7)
    team: int = betterproto.uint32_field(8)
    score: int = betterproto.uint32_field(9)
    ping: int = betterproto.uint32_field(10)
    flags: int = betterproto.uint32_field(11)
    display_rating: int = betterproto.uint32_field(12)
    display_rating_change: int = betterproto.int32_field(13)
    rank: int = betterproto.uint32_field(14)
    classes_played: int = betterproto.uint32_field(15)
    kills: int = betterproto.uint32_field(16)
    deaths: int = betterproto.uint32_field(17)
    damage: int = betterproto.uint32_field(18)
    healing: int = betterproto.uint32_field(19)
    support: int = betterproto.uint32_field(20)
    score_medal: int = betterproto.uint32_field(21)
    kills_medal: int = betterproto.uint32_field(22)
    damage_medal: int = betterproto.uint32_field(23)
    healing_medal: int = betterproto.uint32_field(24)
    support_medal: int = betterproto.uint32_field(25)
    map_index: int = betterproto.uint32_field(26)
    winning_team: int = betterproto.uint32_field(27)


@dataclass(eq=False, repr=False)
class RequestMatchMakerStats(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DataCenterPopulation(betterproto.Message):
    name: str = betterproto.string_field(1)
    health_ratio: float = betterproto.float_field(2)


@dataclass(eq=False, repr=False)
class MatchGroupDataCenterPopulation(betterproto.Message):
    data_center_population: "list[DataCenterPopulation]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class MatchMakerStatsResponse(betterproto.Message):
    map_count: list[int] = betterproto.uint32_field(1)
    matchgroup_data_center_population: "list[MatchGroupDataCenterPopulation]" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class MatchHistoryLoad(betterproto.Message):
    match_group: "EtfMatchGroup" = betterproto.enum_field(1)


@dataclass(eq=False, repr=False)
class DataCenterPingUpdate(betterproto.Message):
    pingdata: "list[DataCenterPingUpdatePingEntry]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DataCenterPingUpdatePingEntry(betterproto.Message):
    name: str = betterproto.string_field(1)
    ping: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class NewMatchForLobbyRequest(betterproto.Message):
    current_match_id: int = betterproto.uint64_field(1)
    next_map_id: int = betterproto.uint32_field(2)
    lobby_id: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class NewMatchForLobbyResponse(betterproto.Message):
    success: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class ChangeMatchPlayerTeamsRequest(betterproto.Message):
    match_id: int = betterproto.uint64_field(1)
    lobby_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class ChangeMatchPlayerTeamsResponse(betterproto.Message):
    success: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class QuestCompleteDebug(betterproto.Message):
    quest_id: int = betterproto.uint64_field(1)
    points_type: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class QuestMapDebug(betterproto.Message):
    reset_operation: int = betterproto.uint32_field(1)
    give_credit: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class QuestMapPurchaseReward(betterproto.Message):
    store_item_defindex: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class QuestResponse(betterproto.Message):
    success: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class SetDisablePartyQuestProgress(betterproto.Message):
    state: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class QuestProgressReport(betterproto.Message):
    quest_id: int = betterproto.uint64_field(1)
    star_0_earned: bool = betterproto.bool_field(2)
    star_1_earned: bool = betterproto.bool_field(3)
    star_2_earned: bool = betterproto.bool_field(4)
    items_earned: list[int] = betterproto.uint64_field(5)
    reward_credits_earned: int = betterproto.uint32_field(6)
    contract_completed: bool = betterproto.bool_field(7)


@dataclass(eq=False, repr=False)
class ConsumePaintkit(betterproto.Message):
    source_id: int = betterproto.fixed64_field(1)
    target_defindex: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class PainkitDevGrant(betterproto.Message):
    paintkit_defindex: int = betterproto.uint32_field(1)
    wear: float = betterproto.float_field(2)
    item_defindex: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class QuestStrangeEvent(betterproto.Message):
    owner_account_id: int = betterproto.uint32_field(1)
    scorer_account_id: int = betterproto.uint32_field(2)
    quest_id: int = betterproto.uint64_field(3)
    strange_event_id: int = betterproto.uint32_field(4)
    score: int = betterproto.uint32_field(5)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Final

from ._const import URL
from .models import CDNAsset
from .utils import DateTime

if TYPE_CHECKING:
    from datetime import datetime

    from .app import App, PartialApp
    from .enums import Language
    from .protobufs import user_stats
    from .state import ConnectionState
    from .types import achievement

__all__ = (
    "AppStat",
    "UserAppStat",
    "AppAchievement",
    "AppStatAchievement",
    "UserAppAchievement",
    "AppStats",
    "UserAppStats",
)


class StatType:  # https://partner.steamgames.com/doc/api/steam_api#ESteamUserStatType
    Int: Final = "1"
    Float: Final = "2"
    AverageRate: Final = "3"
    Achievement: Final = "4"
    GroupAchievement: Final = "5"


@dataclass(slots=True)
class Stat:
    name: str
    """The name of the stat."""
    display_name: str
    """The display name of the stat."""
    app: PartialApp
    """The app this stat is for."""


@dataclass(slots=True)
class AppStat(Stat):
    """Represents a stat for an app."""

    default_value: int
    """The value of the stat."""


@dataclass(slots=True)
class UserAppStat(Stat):
    """Represents a stat for a user in an app."""

    increment_only: bool
    """Whether the stat can only be incremented."""
    value: int = field(init=False)
    """The value of the stat."""


@dataclass(slots=True)
class Achievement:
    name: str
    """The name of the achievement."""
    app: PartialApp
    """The app this achievement is for."""


@dataclass(slots=True)
class AppAchievement(Achievement):
    """Represents an achievement for an app."""

    display_name: str
    """The display name of the achievement."""
    description: str
    """The description of the achievement."""
    icon: CDNAsset
    """The icon of the achievement."""
    icon_gray: CDNAsset
    """The gray icon of the achievement."""
    hidden: bool
    """Whether the achievement is hidden from the app's achievement list."""
    global_percent_unlocked: float
    """The percentage of the players that have this achievement."""


@dataclass(slots=True)
class UserNewsAchievement(Achievement):
    """Represents an achievement for an app."""

    display_name: str
    """The display name of the achievement."""
    description: str
    """The description of the achievement."""
    icon: CDNAsset
    """The icon of the achievement."""
    hidden: bool
    """Whether the achievement is hidden from the app's achievement list."""
    percent_unlocked: float
    """The percentage of the players that have this achievement."""


@dataclass(slots=True)
class AppStatAchievement(Achievement):
    """Represents an achievement for an app."""

    display_name: str
    """The localised name of the achievement."""
    description: str
    """The description of the achievement."""
    default_value: int
    """The default value of the stat."""
    icon: CDNAsset
    """The icon of the achievement."""
    icon_gray: CDNAsset
    """The gray icon of the achievement."""
    hidden: bool
    """Whether the achievement is hidden from the app's achievement list."""


@dataclass(slots=True)
class UserAppAchievement(Achievement):
    """Represents an achievement for a user in an app."""

    display_name: str
    """The localised name of the achievement."""
    min: int | None
    """The minimum value of the stat."""
    max: int | None
    """The maximum value of the achievement."""
    icon: CDNAsset
    """The icon of the achievement."""
    icon_gray: CDNAsset
    """The gray icon of the achievement."""
    hidden: bool
    """Whether the achievement is hidden from the app's achievement list."""
    description: str
    """The description of the achievement."""
    unlocked_at: datetime | None = field(init=False)
    """The time the achievement was unlocked at."""


# This currently isn't useful cause it doesn't include the internal name or the time the player unlocked the
# achievement
# @dataclass(slots=True)
# class UserAchievement(Achievement):
#     description: str
#     """The description of the achievement."""
#     icon: CDNAsset
#     """The icon of the achievement."""
#     icon_gray: CDNAsset
#     """The gray icon of the achievement."""
#     hidden: bool
#     """Whether the achievement is hidden from the app's achievement list."""
#     global_percent_unlocked: float
#     """The percentage of the players that have this achievement."""


class AppStats:
    """Represents a collection of stats and achievements for an app."""

    def __init__(self, state: ConnectionState, app: App, data: achievement.AppAppStats) -> None:
        from .app import PartialApp

        self.app = PartialApp(state, id=app.id, name=data["gameName"])
        """The app these stats and achievement are for."""
        self.version = int(data["gameVersion"])
        """The version of the schema used to generate the stats and achievements."""
        self.achievements = [
            AppStatAchievement(
                achievement["name"],
                self.app,
                achievement["displayName"],
                achievement["description"],
                achievement["defaultvalue"],
                CDNAsset(state, achievement["icon"]),
                CDNAsset(state, achievement["icongray"]),
                bool(achievement["hidden"]),
            )
            for achievement in data["availableGameStats"].get("achievements", [])
        ]
        """The achievements for the app."""
        self.stats = [
            AppStat(
                stat["name"],
                stat["displayName"],
                self.app,
                stat["defaultvalue"],
            )
            for stat in data["availableGameStats"].get("stats", [])
        ]
        """The stats for the app."""

    def __repr__(self) -> str:
        return f"<AppStats app={self.app!r}>"


class UserAppStats:
    """Represents a collection of stats and achievements for a user in an app."""

    def __init__(
        self,
        state: ConnectionState,
        app: App,
        msg: user_stats.CMsgClientGetUserStatsResponse,
        data: achievement.UserAppStats,
        language: Language,
    ) -> None:
        from .app import PartialApp

        self.app = PartialApp(state, id=app.id, name=data["gamename"])
        """The app these stats and achievement are for."""
        self.version = int(data["version"])
        """The version of the schema used to generate the stats and achievements."""
        achievements = {
            int(achievement_["id"]): [
                UserAppAchievement(
                    achievement["name"],
                    self.app,
                    (display := achievement["display"])["name"].get(language.api_name, ""),
                    (
                        int(progress["min_value"])
                        if (progress := achievement.get("progress")) and "min_value" in progress
                        else None
                    ),
                    int(progress["max_value"]) if progress and "max_value" in progress else None,
                    CDNAsset(
                        state,
                        f"{URL.CDN}/steamcommunity/public/images/apps/{app.id}/{display['icon']}",
                    ),
                    CDNAsset(
                        state,
                        f"{URL.CDN}/steamcommunity/public/images/apps/{app.id}/{display['icon_gray']}",
                    ),
                    bool(int(display["hidden"])),
                    display["desc"].get(language.api_name, ""),
                )
                for achievement in achievement_["bits"].values()
            ]
            for achievement_ in data["stats"].values()
            if achievement_["type"] == StatType.Achievement
        }
        stats = {
            int(stat["id"]): UserAppStat(
                stat["name"],
                stat["display"].get("name", ""),
                self.app,
                bool(stat.get("incrementonly", False)),
            )
            for stat in data["stats"].values()
            if stat["type"] == StatType.Int
        }

        self.stats: list[UserAppStat] = []
        """The stats for the app."""
        self.achievements: list[UserAppAchievement] = []
        """The achievements for the app."""

        for stat in msg.stats:
            try:
                app_stat = stats[stat.stat_id]
            except KeyError:
                continue  # not really sure why this happens, they are in achievements for some reason
            app_stat.value = stat.stat_value
            self.stats.append(app_stat)

        # this is a contender for the worse steam API data representation there is
        # achievements are chunked into blocks of 32 and give each block an id, then the unlock_time are a list of the same length
        # and you have to zip them together to get the correct unlock time for each achievement after finding it in
        # doubly nested dict of the schema. Why??????
        for achievement in msg.achievement_blocks:
            for app_achievement, unlock_time in zip(achievements[achievement.achievement_id], achievement.unlock_time):
                app_achievement.unlocked_at = DateTime.from_timestamp(unlock_time) if unlock_time else None
                self.achievements.append(app_achievement)

    def __repr__(self) -> str:
        return f"<UserAppStats app={self.app!r}>"



from __future__ import annotations

from typing import TYPE_CHECKING

from ....protobufs.msg import GCMessage, GCProtobufMessage
from ....utils import StructIO
from ..enums import EMsg

if TYPE_CHECKING:
    from ....types.id import ID64, AssetID

# some custom messages to make things a lot easier decoding/encoding wise


class CraftRequest(GCMessage, msg=EMsg.Craft):
    recipe: int
    items: list[AssetID]

    def __bytes__(self) -> bytes:
        with StructIO() as io:
            io.write_struct("<hh", self.recipe, len(self.items))
            io.write_struct(f"<{len(self.items)}Q", *self.items)

            return io.buffer


class CraftResponse(GCMessage, msg=EMsg.CraftResponse):
    recipe_id: int
    ids: tuple[AssetID, ...]
    being_used: bool

    def parse(self, data: bytes, /) -> CraftResponse:  # type: ignore
        with StructIO(data) as io:
            self.recipe_id = io.read_i16()
            _ = io.read_u32()  # always 0 in mckay's experience
            id_count = io.read_i16()
            self.ids = io.read_struct(f"<{id_count}Q")
            self.being_used = False

        return self


class SetItemStyleRequest(GCMessage, msg=EMsg.SetItemStyle):
    item_id: AssetID
    style: int

    def __bytes__(self) -> bytes:
        with StructIO() as io:
            io.write_u64(self.item_id)
            io.write_u32(self.style)

            return io.buffer


class DeleteItemRequest(GCMessage, msg=EMsg.Delete):
    item_id: AssetID


class WrapItemRequest(GCMessage, msg=EMsg.GiftWrapItem):
    wrapping_paper_id: AssetID
    item_id: AssetID


class UnwrapItemRequest(GCMessage, msg=EMsg.UnwrapGiftRequest):
    gift_id: AssetID


class DeliverGiftRequest(GCMessage, msg=EMsg.DeliverGift):
    user_id64: ID64
    gift_id: AssetID


class OpenCrateRequest(GCMessage, msg=EMsg.UnlockCrate):
    key_id: AssetID
    crate_id: AssetID


class CacheSubscribedCheck(GCProtobufMessage, msg=EMsg.SOCacheSubscriptionCheck):
    pass



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from typing import TypedDict


class AddPhoneNumber(TypedDict):
    success: bool
    email_confirmation: bool
    error_text: str
    fatal: bool



# Heavily based on https://github.com/ValvePython/csgo/blob/master/csgo/sharecode.py

import itertools
import re
import zlib
from array import array
from collections.abc import Mapping
from dataclasses import dataclass
from functools import reduce
from typing import Final, List, NamedTuple, cast

CHARS = cast(
    Mapping[str, int], dict(zip("ABCDEFGHJKLMNOPQRSTUVWXYZabcdefhijkmnopqrstuvwxyz23456789", itertools.count()))
)


def _swap_endianness(number: int, ns: list[int] = list(range(0, 144, 8)), /) -> int:
    return reduce(lambda result, n: (result << 8) + ((number >> n) & 0xFF), ns, 0)


SHARE_CODE_RE: Final = re.compile(rf"^(CSGO)?(-?[{''.join(CHARS)}]{{5}}){{5}}$")


class ShareCode(NamedTuple):
    match_id: int
    outcome_id: int
    token: int


def decode_sharecode(code: str) -> ShareCode:
    """Decodes a match share code.

    Returns
    -------
    .. source:: ShareCode
    """
    if SHARE_CODE_RE.match(code) is None:
        raise ValueError("Invalid share code")

    full_code = _swap_endianness(
        reduce(
            lambda bits, char: (bits * len(CHARS)) + CHARS[char],
            code.removeprefix("CSGO-").replace("-", "")[::-1],
            0,
        )
    )

    return ShareCode(
        full_code & 0xFFFFFFFFFFFFFFFF,
        full_code >> 64 & 0xFFFFFFFFFFFFFFFF,
        full_code >> 128 & 0xFFFF,
    )


@dataclass
class Keychain:
    slot: int
    sticker_id: int
    pattern: int


def create_inspect_link(def_index: int, rarity: int, quality: int, paint_seed: int, keychains: list[Keychain]) -> str:
    # Create proto-like structure as bytes
    proto = bytearray([def_index & 0xFF, rarity & 0xFF, quality & 0xFF, paint_seed & 0xFF])

    for chain in keychains:
        proto.extend([chain.slot & 0xFF, chain.sticker_id & 0xFF, chain.pattern & 0xFF])

    # Create final buffer
    buffer = bytearray([0]) + proto

    # Calculate CRC
    crc = zlib.crc32(buffer) & 0xFFFFFFFF
    xored_crc = (crc & 0xFFFF) ^ (len(proto) * crc)

    # Add checksum
    checksum = array("I", [xored_crc]).tobytes()
    final_buffer = buffer + checksum

    # Format URL
    hex_string = "".join(f"{b:02X}" for b in final_buffer)
    return f"steam://rungame/730/76561202255233023/+csgo_econ_action_preview {hex_string}"


def generate_keychain_inspect_url(pattern: int, keychain_id: int) -> str:
    keychains = [Keychain(slot=0, sticker_id=keychain_id, pattern=pattern)]
    return create_inspect_link(1355, 4, 4, pattern, keychains)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_store.steamclient.proto, steammessages_storebrowse.steamclient.proto, service_storequery.proto
# plugin: python-betterproto
# Last updated 27/02/2022

from dataclasses import dataclass

import betterproto

from .msg import UnifiedMessage


class EContentDescriptorId(betterproto.Enum):
    FrequentNudityOrSexualContent = 1
    FrequentViolenceOrGore = 2
    StrongSexualContent = 3
    UNUSED_4 = 4
    AnyMatureContent = 5


class ESteamDeckCompatibilityCategory(betterproto.Enum):
    Unknown = 0
    Unsupported = 1
    Playable = 2
    Verified = 3


class EStoreItemType(betterproto.Enum):
    Invalid = -1
    App = 0
    Package = 1
    Bundle = 2
    MicroTransaction = 3
    Tag = 4
    Creator = 5
    HubCategory = 6


class EStoreAppType(betterproto.Enum):
    Game = 0
    Demo = 1
    Mod = 2
    Movie = 3
    DLC = 4
    Guide = 5
    Software = 6
    Video = 7
    Series = 8
    Episode = 9
    Hardware = 10
    Music = 11
    Beta = 12
    Tool = 13
    Advertising = 14


class EUserReviewScore(betterproto.Enum):
    NONE = 0
    OverwhelminglyNegative = 1
    VeryNegative = 2
    Negative = 3
    MostlyNegative = 4
    Mixed = 5
    MostlyPositive = 6
    Positive = 7
    VeryPositive = 8
    OverwhelminglyPositive = 9


class EStoreCategoryType(betterproto.Enum):
    Category = 0
    SupportedPlayers = 1
    Feature = 2
    ControllerSupport = 3
    CloudGaming = 4
    MAX = 5


class RegisterCDKeyRequest(UnifiedMessage, um_name="Store.RegisterCDKey"):
    activation_code: str = betterproto.string_field(1)
    purchase_platform: int = betterproto.int32_field(2)
    is_request_from_client: bool = betterproto.bool_field(3)


class RegisterCDKeyResponse(UnifiedMessage, um_name="Store.RegisterCDKey"):
    eresult: int = betterproto.int32_field(1)
    purchase_receipt_info: "PurchaseReceiptInfo" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class PurchaseReceiptInfo(betterproto.Message):
    transactionid: int = betterproto.uint64_field(1)
    packageid: int = betterproto.uint32_field(2)
    purchase_status: int = betterproto.uint32_field(3)
    result_detail: int = betterproto.uint32_field(4)
    transaction_time: int = betterproto.uint32_field(5)
    payment_method: int = betterproto.uint32_field(6)
    base_price: int = betterproto.uint64_field(7)
    total_discount: int = betterproto.uint64_field(8)
    tax: int = betterproto.uint64_field(9)
    shipping: int = betterproto.uint64_field(10)
    currency_code: int = betterproto.uint32_field(11)
    country_code: str = betterproto.string_field(12)
    error_headline: str = betterproto.string_field(13)
    error_string: str = betterproto.string_field(14)
    error_link_text: str = betterproto.string_field(15)
    error_link_url: str = betterproto.string_field(16)
    error_appid: int = betterproto.uint32_field(17)
    line_items: "list[PurchaseReceiptInfoLineItem]" = betterproto.message_field(18)


@dataclass(eq=False, repr=False)
class PurchaseReceiptInfoLineItem(betterproto.Message):
    packageid: int = betterproto.uint32_field(1)
    appid: int = betterproto.uint32_field(2)
    line_item_description: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class StoreItemId(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    packageid: int = betterproto.uint32_field(2)
    bundleid: int = betterproto.uint32_field(3)
    tagid: int = betterproto.uint32_field(4)
    creatorid: int = betterproto.uint32_field(5)
    hubcategoryid: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class StoreBrowseContext(betterproto.Message):
    language: str = betterproto.string_field(1)
    elanguage: int = betterproto.int32_field(2)
    country_code: str = betterproto.string_field(3)
    steam_realm: int = betterproto.int32_field(4)


@dataclass(eq=False, repr=False)
class StoreBrowseItemDataRequest(betterproto.Message):
    include_assets: bool = betterproto.bool_field(1)
    include_release: bool = betterproto.bool_field(2)
    include_platforms: bool = betterproto.bool_field(3)
    include_all_purchase_options: bool = betterproto.bool_field(4)
    include_screenshots: bool = betterproto.bool_field(5)
    include_trailers: bool = betterproto.bool_field(6)
    include_ratings: bool = betterproto.bool_field(7)
    include_tag_count: int = betterproto.int32_field(8)
    include_reviews: bool = betterproto.bool_field(9)
    include_basic_info: bool = betterproto.bool_field(10)
    include_supported_languages: bool = betterproto.bool_field(11)


class GetItemsRequest(UnifiedMessage, um_name="StoreBrowse.GetItems"):
    ids: list[StoreItemId] = betterproto.message_field(1)
    context: "StoreBrowseContext" = betterproto.message_field(2)
    data_request: "StoreBrowseItemDataRequest" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class StoreGameRating(betterproto.Message):
    type: str = betterproto.string_field(1)
    rating: str = betterproto.string_field(2)
    descriptors: list[str] = betterproto.string_field(3)
    interactive_elements: str = betterproto.string_field(4)
    required_age: int = betterproto.int32_field(10)
    use_age_gate: bool = betterproto.bool_field(11)
    image_url: str = betterproto.string_field(20)
    image_target: str = betterproto.string_field(21)


@dataclass(eq=False, repr=False)
class StoreItem(betterproto.Message):
    item_type: EStoreItemType = betterproto.enum_field(1)
    id: int = betterproto.uint32_field(2)
    success: int = betterproto.uint32_field(3)
    visible: bool = betterproto.bool_field(4)
    unvailable_for_country_restriction: bool = betterproto.bool_field(5)
    name: str = betterproto.string_field(6)
    store_url_path: str = betterproto.string_field(7)
    appid: int = betterproto.uint32_field(9)
    type: EStoreAppType = betterproto.enum_field(10)
    included_types: list[EStoreAppType] = betterproto.enum_field(11)
    included_appids: list[int] = betterproto.uint32_field(12)
    is_free: bool = betterproto.bool_field(13)
    is_early_access: bool = betterproto.bool_field(14)
    related_items: "StoreItemRelatedItems" = betterproto.message_field(15)
    content_descriptorids: list[EContentDescriptorId] = betterproto.enum_field(20)
    tagids: list[int] = betterproto.uint32_field(21)
    categories: "StoreItemCategories" = betterproto.message_field(22)
    reviews: "StoreItemReviews" = betterproto.message_field(23)
    basic_info: "StoreItemBasicInfo" = betterproto.message_field(24)
    tags: "list[StoreItemTag]" = betterproto.message_field(25)
    assets: "StoreItemAssets" = betterproto.message_field(30)
    release: "StoreItemReleaseInfo" = betterproto.message_field(31)
    platforms: "StoreItemPlatforms" = betterproto.message_field(32)
    game_rating: StoreGameRating = betterproto.message_field(33)
    best_purchase_option: "StoreItemPurchaseOption" = betterproto.message_field(40)
    purchase_options: "list[StoreItemPurchaseOption]" = betterproto.message_field(41)
    accessories: "list[StoreItemPurchaseOption]" = betterproto.message_field(42)
    screenshots: "StoreItemScreenshots" = betterproto.message_field(50)
    trailers: "StoreItemTrailers" = betterproto.message_field(51)
    supported_languages: "list[StoreItemSupportedLanguage]" = betterproto.message_field(52)
    store_url_path_override: str = betterproto.string_field(53)
    free_weekend: "StoreItemFreeWeekend" = betterproto.message_field(54)
    unlisted: bool = betterproto.bool_field(55)


@dataclass(eq=False, repr=False)
class StoreItemRelatedItems(betterproto.Message):
    parent_appid: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class StoreItemCategories(betterproto.Message):
    supported_player_categoryids: list[int] = betterproto.uint32_field(2)
    feature_categoryids: list[int] = betterproto.uint32_field(3)
    controller_categoryids: list[int] = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class StoreItemReviews(betterproto.Message):
    summary_filtered: "StoreItemReviewsStoreReviewSummary" = betterproto.message_field(1)
    summary_unfiltered: "StoreItemReviewsStoreReviewSummary" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class StoreItemReviewsStoreReviewSummary(betterproto.Message):
    review_count: int = betterproto.uint32_field(1)
    percent_positive: int = betterproto.int32_field(2)
    review_score: EUserReviewScore = betterproto.enum_field(3)
    review_score_label: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class StoreItemBasicInfo(betterproto.Message):
    short_description: str = betterproto.string_field(1)
    publishers: "list[StoreItemBasicInfoCreatorHomeLink]" = betterproto.message_field(2)
    developers: "list[StoreItemBasicInfoCreatorHomeLink]" = betterproto.message_field(3)
    franchises: "list[StoreItemBasicInfoCreatorHomeLink]" = betterproto.message_field(4)
    capsule_headline: str = betterproto.string_field(5)


@dataclass(eq=False, repr=False)
class StoreItemBasicInfoCreatorHomeLink(betterproto.Message):
    name: str = betterproto.string_field(1)
    creator_clan_account_id: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class StoreItemTag(betterproto.Message):
    tagid: int = betterproto.uint32_field(1)
    weight: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class StoreItemAssets(betterproto.Message):
    asset_url_format: str = betterproto.string_field(1)
    main_capsule: str = betterproto.string_field(2)
    small_capsule: str = betterproto.string_field(3)
    header: str = betterproto.string_field(4)
    package_header: str = betterproto.string_field(5)
    page_background: str = betterproto.string_field(6)
    hero_capsule: str = betterproto.string_field(7)
    hero_capsule_2_x: str = betterproto.string_field(8)
    library_capsule: str = betterproto.string_field(9)
    library_capsule_2_x: str = betterproto.string_field(10)
    library_hero: str = betterproto.string_field(11)
    library_hero_2_x: str = betterproto.string_field(12)
    community_icon: str = betterproto.string_field(13)


@dataclass(eq=False, repr=False)
class StoreItemReleaseInfo(betterproto.Message):
    steam_release_date: int = betterproto.uint32_field(1)
    original_release_date: int = betterproto.uint32_field(2)
    original_steam_release_date: int = betterproto.uint32_field(3)
    is_coming_soon: bool = betterproto.bool_field(4)
    is_preload: bool = betterproto.bool_field(5)
    custom_release_date_message: str = betterproto.string_field(6)
    is_abridged_release_date: bool = betterproto.bool_field(7)
    coming_soon_display: str = betterproto.string_field(8)
    is_early_access: bool = betterproto.bool_field(10)
    mac_release_date: int = betterproto.uint32_field(20)
    linux_release_date: int = betterproto.uint32_field(21)


@dataclass(eq=False, repr=False)
class StoreItemPlatforms(betterproto.Message):
    windows: bool = betterproto.bool_field(1)
    mac: bool = betterproto.bool_field(2)
    steamos_linux: bool = betterproto.bool_field(3)
    vr_support: "StoreItemPlatformsVrSupport" = betterproto.message_field(10)
    steam_deck_compat_category: ESteamDeckCompatibilityCategory = betterproto.enum_field(11)


@dataclass(eq=False, repr=False)
class StoreItemPlatformsVrSupport(betterproto.Message):
    vrhmd: bool = betterproto.bool_field(1)
    vrhmd_only: bool = betterproto.bool_field(2)
    htc_vive: bool = betterproto.bool_field(40)
    oculus_rift: bool = betterproto.bool_field(41)
    windows_mr: bool = betterproto.bool_field(42)
    valve_index: bool = betterproto.bool_field(43)


@dataclass(eq=False, repr=False)
class StoreItemPurchaseOption(betterproto.Message):
    packageid: int = betterproto.int32_field(1)
    bundleid: int = betterproto.int32_field(2)
    purchase_option_name: str = betterproto.string_field(3)
    final_price_in_cents: int = betterproto.int64_field(5)
    original_price_in_cents: int = betterproto.int64_field(6)
    user_final_price_in_cents: int = betterproto.int64_field(7)
    formatted_final_price: str = betterproto.string_field(8)
    formatted_original_price: str = betterproto.string_field(9)
    discount_pct: int = betterproto.int32_field(10)
    user_discount_pct: int = betterproto.int32_field(11)
    bundle_discount_pct: int = betterproto.int32_field(12)
    active_discounts: "list[StoreItemPurchaseOptionDiscount]" = betterproto.message_field(20)
    user_active_discounts: "list[StoreItemPurchaseOptionDiscount]" = betterproto.message_field(21)
    inactive_discounts: "list[StoreItemPurchaseOptionDiscount]" = betterproto.message_field(22)
    user_can_purchase: bool = betterproto.bool_field(30)
    user_can_purchase_as_gift: bool = betterproto.bool_field(31)
    is_commercial_license: bool = betterproto.bool_field(40)
    should_suppress_discount_pct: bool = betterproto.bool_field(41)


@dataclass(eq=False, repr=False)
class StoreItemPurchaseOptionDiscount(betterproto.Message):
    discount_amount: int = betterproto.int64_field(1)
    discount_description: str = betterproto.string_field(2)
    discount_end_date: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class StoreItemScreenshots(betterproto.Message):
    all_ages_screenshots: "list[StoreItemScreenshotsScreenshot]" = betterproto.message_field(2)
    mature_content_screenshots: "list[StoreItemScreenshotsScreenshot]" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class StoreItemScreenshotsScreenshot(betterproto.Message):
    filename: str = betterproto.string_field(1)
    ordinal: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class StoreItemTrailers(betterproto.Message):
    highlights: "list[StoreItemTrailersTrailer]" = betterproto.message_field(1)
    other_trailers: "list[StoreItemTrailersTrailer]" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class StoreItemTrailersVideoSource(betterproto.Message):
    filename: str = betterproto.string_field(1)
    type: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class StoreItemTrailersTrailer(betterproto.Message):
    trailer_name: str = betterproto.string_field(1)
    trailer_url_format: str = betterproto.string_field(2)
    trailer_480_p: list[StoreItemTrailersVideoSource] = betterproto.message_field(3)
    trailer_max: list[StoreItemTrailersVideoSource] = betterproto.message_field(4)
    microtrailer: list[StoreItemTrailersVideoSource] = betterproto.message_field(5)
    screenshot_medium: str = betterproto.string_field(10)
    screenshot_full: str = betterproto.string_field(11)
    trailer_base_id: int = betterproto.int32_field(12)


@dataclass(eq=False, repr=False)
class StoreItemSupportedLanguage(betterproto.Message):
    elanguage: int = betterproto.int32_field(1)
    supported: bool = betterproto.bool_field(2)
    full_audio: bool = betterproto.bool_field(3)
    subtitles: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class StoreItemFreeWeekend(betterproto.Message):
    start_time: int = betterproto.uint32_field(1)
    end_time: int = betterproto.uint32_field(2)
    text: str = betterproto.string_field(3)


class GetItemsResponse(UnifiedMessage, um_name="StoreBrowse.GetItems"):
    store_items: list[StoreItem] = betterproto.message_field(1)


class GetStoreCategoriesRequest(UnifiedMessage, um_name="StoreBrowse.GetStoreCategories"):
    language: str = betterproto.string_field(1)
    elanguage: int = betterproto.int32_field(2)


class GetStoreCategoriesResponse(UnifiedMessage, um_name="StoreBrowse.GetStoreCategories"):
    categories: "list[GetStoreCategoriesResponseCategory]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetStoreCategoriesResponseCategory(betterproto.Message):
    categoryid: int = betterproto.uint32_field(1)
    type: EStoreCategoryType = betterproto.enum_field(2)
    internal_name: str = betterproto.string_field(3)
    display_name: str = betterproto.string_field(4)
    image_url: str = betterproto.string_field(5)
    show_in_search: bool = betterproto.bool_field(6)


@dataclass(eq=False, repr=False)
class StorePageFilter(betterproto.Message):
    @dataclass(eq=False, repr=False)
    class SalePageFilter(betterproto.Message):
        sale_tagid: int = betterproto.uint32_field(1)

    sale_filter: SalePageFilter = betterproto.message_field(1)

    @dataclass(eq=False, repr=False)
    class ContentHubFilter(betterproto.Message):
        @dataclass(eq=False, repr=False)
        class OptInInfo(betterproto.Message):
            name: str = betterproto.string_field(1)
            optin_tagid: int = betterproto.uint32_field(2)
            prune_tagid: int = betterproto.uint32_field(3)
            optin_only: bool = betterproto.bool_field(4)

        hub_type: str = betterproto.string_field(1)
        hub_category: str = betterproto.string_field(2)
        hub_tagid: int = betterproto.uint32_field(3)
        discount_filter: int = betterproto.int32_field(4)
        optin: "OptInInfo" = betterproto.message_field(5)

    content_hub_filter: ContentHubFilter = betterproto.message_field(2)

    @dataclass(eq=False, repr=False)
    class StoreFilter(betterproto.Message):
        filter_json: str = betterproto.string_field(1)
        cache_key: str = betterproto.string_field(2)

    store_filters: list[StoreFilter] = betterproto.message_field(3)


class GetDLCForAppsRequest(UnifiedMessage, um_name="StoreBrowse.GetDLCForApps"):
    context: StoreBrowseContext = betterproto.message_field(1)
    store_page_filter: StorePageFilter = betterproto.message_field(2)
    appids: list[StoreItemId] = betterproto.message_field(3)
    steamid: int = betterproto.uint64_field(4)


@dataclass(eq=False, repr=False)
class GetDLCForAppsResponseDLCData(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    parentappid: int = betterproto.uint32_field(2)
    release_date: int = betterproto.uint32_field(3)
    coming_soon: bool = betterproto.bool_field(4)
    price: int = betterproto.int64_field(5)
    discount: int = betterproto.uint32_field(6)
    free: bool = betterproto.bool_field(7)


@dataclass(eq=False, repr=False)
class GetDLCForAppsResponsePlaytimeForApp(betterproto.Message):
    appid: int = betterproto.uint32_field(1)
    playtime: int = betterproto.uint32_field(2)
    last_played: int = betterproto.uint32_field(3)


class GetDLCForAppsResponse(UnifiedMessage, um_name="StoreBrowse.GetDLCForApps"):
    dlc_data: "list[GetDLCForAppsResponseDLCData]" = betterproto.message_field(1)
    playtime: "list[GetDLCForAppsResponsePlaytimeForApp]" = betterproto.message_field(2)


class QueryRequest(UnifiedMessage, um_name="StoreQuery.Query"):
    query_name: str = betterproto.string_field(1)

    @dataclass(eq=False, repr=False)
    class StoreQueryParams(betterproto.Message):
        start: int = betterproto.int32_field(1)
        count: int = betterproto.int32_field(2)
        sort: int = betterproto.int32_field(10)
        filters: "StoreQueryFilters" = betterproto.message_field(20)

    query: StoreQueryParams = betterproto.message_field(2)
    context: StoreBrowseContext = betterproto.message_field(3)
    data_request: StoreBrowseItemDataRequest = betterproto.message_field(4)
    override_country_code: str = betterproto.string_field(5)


class QueryResponse(UnifiedMessage, um_name="StoreQuery.Query"):
    metadata: "StoreQueryResultMetadata" = betterproto.message_field(1)
    ids: list[StoreItemId] = betterproto.message_field(2)
    store_items: list[StoreItem] = betterproto.message_field(3)


class SearchSuggestionsRequest(UnifiedMessage, um_name="StoreQuery.SearchSuggestions"):
    query_name: str = betterproto.string_field(1)
    context: StoreBrowseContext = betterproto.message_field(2)
    search_term: str = betterproto.string_field(3)
    max_results: int = betterproto.uint32_field(4)
    filters: "StoreQueryFilters" = betterproto.message_field(5)
    data_request: StoreBrowseItemDataRequest = betterproto.message_field(6)
    use_spellcheck: bool = betterproto.bool_field(7)
    search_tags: bool = betterproto.bool_field(8)
    search_creators: bool = betterproto.bool_field(9)
    prefilter_creators: bool = betterproto.bool_field(10)


class SearchSuggestionsResponse(UnifiedMessage, um_name="StoreQuery.SearchSuggestions"):
    metadata: "StoreQueryResultMetadata" = betterproto.message_field(1)
    ids: list[StoreItemId] = betterproto.message_field(2)
    store_items: list[StoreItem] = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class StoreQueryFilters(betterproto.Message):
    released_only: bool = betterproto.bool_field(1)
    coming_soon_only: bool = betterproto.bool_field(2)

    @dataclass(eq=False, repr=False)
    class TypeFilters(betterproto.Message):
        include_apps: bool = betterproto.bool_field(1)
        include_packages: bool = betterproto.bool_field(2)
        include_bundles: bool = betterproto.bool_field(3)
        include_games: bool = betterproto.bool_field(10)
        include_demos: bool = betterproto.bool_field(11)
        include_mods: bool = betterproto.bool_field(12)
        include_dlc: bool = betterproto.bool_field(13)
        include_software: bool = betterproto.bool_field(14)
        include_video: bool = betterproto.bool_field(15)
        include_hardware: bool = betterproto.bool_field(16)
        include_series: bool = betterproto.bool_field(17)
        include_music: bool = betterproto.bool_field(18)

    type_filters: TypeFilters = betterproto.message_field(3)

    @dataclass(eq=False, repr=False)
    class TagFilter(betterproto.Message):
        tagids: list[int] = betterproto.int32_field(1)

    tagids_must_match: list[TagFilter] = betterproto.message_field(10)
    tagids_exclude: list[int] = betterproto.uint32_field(11)
    content_descriptors_must_match: list[EContentDescriptorId] = betterproto.enum_field(15)
    content_descriptors_excluded: list[EContentDescriptorId] = betterproto.enum_field(16)
    regional_top_n_sellers: int = betterproto.int32_field(40)
    global_top_n_sellers: int = betterproto.int32_field(41)
    regional_long_term_top_n_sellers: int = betterproto.int32_field(42)
    global_long_term_top_n_sellers: int = betterproto.int32_field(43)
    store_page_filter: StorePageFilter = betterproto.message_field(44)
    parent_appids: list[int] = betterproto.uint32_field(45)


@dataclass(eq=False, repr=False)
class StoreQueryPerResultMetadata(betterproto.Message):
    id: StoreItemId = betterproto.message_field(1)
    score: float = betterproto.double_field(2)
    spellcheck_generated_result: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class StoreQueryResultMetadata(betterproto.Message):
    total_matching_records: int = betterproto.int32_field(1)
    start: int = betterproto.int32_field(2)
    count: int = betterproto.int32_field(3)
    per_result_metadata: list[StoreQueryPerResultMetadata] = betterproto.message_field(4)
    spellcheck_suggestions: list[str] = betterproto.string_field(5)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic

from typing_extensions import Never, TypeVar

from ._const import ReadOnly, impl_eq_via_id
from .abc import Awardable, Commentable
from .types.user import AuthorT

if TYPE_CHECKING:
    from datetime import datetime

    from .reaction import Award, AwardReaction
    from .state import ConnectionState
    from .types.id import CommentID


__all__ = ("Comment",)

OwnerT = TypeVar("OwnerT", bound="Commentable", default="Commentable", covariant=True)


@dataclass(repr=False, slots=True)
@impl_eq_via_id
class Comment(Awardable["CommentID"], Generic[OwnerT, AuthorT]):
    """Represents a comment on a Steam profile.

    .. container:: operations

        .. describe:: x == y

            Checks if two comments are equal.

        .. describe:: hash(x)

            Returns the comment's hash.
    """

    _state: ConnectionState
    id: ReadOnly[CommentID]
    """The comment's ID."""
    content: str
    """The comment's content."""
    created_at: datetime
    """The time the comment was posted at."""
    reactions: list[AwardReaction]
    """The comment's reactions."""
    author: AuthorT
    """The author of the comment."""
    owner: OwnerT
    """The comment sections owner."""

    def __repr__(self) -> str:
        attrs = ("id", "author")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    async def report(self) -> None:
        """Reports the comment."""
        await self._state.report_comment(self.owner, self.id)

    async def delete(self) -> None:
        """Deletes the comment."""
        await self._state.delete_comment(self.owner, self.id)

    if TYPE_CHECKING:
        # @overload
        # async def award(self: Comment[Topic], award: Award) -> None: ...
        async def award(self, award: Award) -> Never: ...
